/*
   This file is copyright (c) 1988 by William Couture.  All rights reserved.
*/


/*
   The functions in this file have been hand optimized after being compiled
   to assembler code.  Do not change this file unless absolutely necessary,
   and then you must re-optimize the functions in ASMFNS.IBM and ASMFNS.ZEN.
*/


#include <compiler.h>
#include <vim.h>
#include <regstruc.h>
#include <opdefs.h>
#include <machine.h>

#define FP_SEG(fp) (*((unsigned int *)&(fp) + 1))

extern union regunion regs;
extern struct tofrom dest;
extern struct addrtrap brkpoints[21];
extern unsigned int brkcounts[21];

extern unsigned int segoffset;
extern unsigned int memok;
extern unsigned int numbreaks;
extern unsigned int datasegment;
extern unsigned int memmap[16];
extern unsigned int dmamap[16];


#ifdef ZENITH
extern unsigned char scratchmem[16386];
#endif

void flags(length)  /* Set the PARITY, SIGN, and ZERO flags.  NOTE:  some
                       instructions leave certain flags undefined.  The
                       simulator simply leaves these flags alone, so the
                       previous state is still available.  Hopefully no one
                       will try to use some undocumented flag conditions. */
   unsigned int length;
   {
   register int i,k;
   int j;

   j = 0;
   k = 1;
   for (i = 0; i < 8; i++)  /* count low 8 bits for parity */
      {
      if (((unsigned int)dest.value) & k)
         j++;
      k <<= 1;
      }
   if (j & 1)
      regs.r.flag &= PARITYOFF;
   else
      regs.r.flag |= PARITY;
   if (length == BYTE)   /* ZERO and SIGN for byte result */
      {
      if (((unsigned char)dest.value) & '\377')
         regs.r.flag &= ZEROOFF;
      else
         regs.r.flag |= ZERO;
      if (((unsigned char)dest.value) & '\200')
         regs.r.flag |= SIGN;
      else
         regs.r.flag &= SIGNOFF;
      }
   else  /* ZERO and SIGN for word result */
      {
      if (((unsigned int)dest.value) & 0xffff)
         regs.r.flag &= ZEROOFF;
      else
         regs.r.flag |= ZERO;
      if (((unsigned int)dest.value) & 0x8000)
         regs.r.flag |= SIGN;
      else
         regs.r.flag &= SIGNOFF;
      }
   }

void ovcarry(value,carryvalue,length)
   unsigned int value,carryvalue,length;
       /* set the CARRY, AUXCARRY, and OVERFLOW flags.  Note something
       interesting that I had to learn:

       When doing subtraction (SUB, SBB, CMP, CMPS, SCAS, etc.) the 8088
       is supposed to do them by 2's complement addition.  According to
       at least one reputable source (The 8086 Book, by Rector and Alexy,
       highly recommended, p.3-243), an example of SBB (subtract with borrow)
       goes something like this:  bl = 0x64, dl = 0x03, carry (borrow) set,
                    0x64 = 0110  0100
        2's comp of 0x03 = 1111  1101
           2's comp of 1 = 1111  1111
                         ------------
                           0110  0000 
                                       carry out of bit 3, set AF to 0
                                       carry out of bit 7, set CY to 0

       This seems fine, until you try something like this:
       bl = 0, dl = 0xff, carry (borrow) set:
                    0x00 = 0000  0000
        2's comp of 0xff = 0000  0001
           2's comp of 1 = 1111  1111
                         ------------
                           0000  0000
                                       carry out of bit 3, set AF to 0
                                       carry out of bit 7, set CY to 0
      Unfortunately, the chip sets AC to 1 and CY to 1.  What is actually
      going on is that that the 1's complement of the negative is being
      taken, and the complement of the carry (borrow) is being taken, to
      make up both the +1 for 2's complement and the carry (borrow).
      In the last example:
                    0x00 = 0000  0000
        1's comp of 0xff = 0000  0000
               comp of 1 =          0    (note the single bit)
                         ------------
                           0000  0000
                                       no carry from bit 3, set AF to 1
                                       no carry from bit 7, set CY to 1 */
   {
   UNSIGNED_LONG temp;
   register unsigned int ovflag,cflag;

/* Note:  This routine is used for both addition and subtraction.  The
   1's complement of the values is passed by the calling routine, so there
   is no need to do so here (and no way to tell addition from subtraction,
   either). */
   temp = dest.oldvalue;
   if ((dest.oldvalue & 0x0f) + (value & 0x0f) + carryvalue > 15)
      regs.r.flag |= AUXCARRY;
   else
      regs.r.flag &= AUXCARRYOFF;  /* AUXCARRY does not depend on BYTE/WORD */
   if (length == BYTE)
      {
      temp &= 0xffL;
      temp += (value & 0xffL) + (carryvalue & 0xffL);
      if (((unsigned int)temp) & 0xff00)
         {
         regs.r.flag |= CARRY;
         cflag = 1;
         }
      else
         {
         regs.r.flag &= CARRYOFF;
         cflag = 0;
         }
      if ((dest.oldvalue & 0x7fL) + (value & 0x7fL) + (carryvalue & 0x7fL) > 0x7f)
         ovflag = 1;
      else
         ovflag = 0;
      if (ovflag == cflag)  /* OVERFLOW set if carry out of sign != carry
                                                                  into sign */
         regs.r.flag &= OVERFLOWOFF;
      else
         regs.r.flag |= OVERFLOW;
      }
   else
      {
      temp &= 0xffffL;
      temp += (((UNSIGNED_LONG)value) & 0xffffL)
                 + (((UNSIGNED_LONG)carryvalue) & 0xffffL);
      if (temp & 0xffff0000L)
         {
         regs.r.flag |= CARRY;
         cflag = 1;
         }
      else
         {
         regs.r.flag &= CARRYOFF;
         cflag = 0;
         }
      if ((dest.oldvalue & 0x7fffL) + (((UNSIGNED_LONG)value) & 0x7fffL) + (((UNSIGNED_LONG)carryvalue) & 0x7fffL) > 0x7fff)
         ovflag = 1;
      else
         ovflag = 0;
      if (ovflag == cflag)
         regs.r.flag &= OVERFLOWOFF;
      else
         regs.r.flag |= OVERFLOW;
      }
   }

void checkaddress1(seg,ofs)  /* check memory, ignore virtual screen */
   unsigned int *seg, *ofs;
   {
   UNSIGNED_LONG temp;
   unsigned int t1,aseg,count;

   aseg = (t1 = *seg + (*ofs >> 4)) >> 12;
   if (numbreaks)
      {
      temp = (((UNSIGNED_LONG)t1) << 4) + (*ofs & 0x000f);
      for (count = 1; count <= numbreaks; count++)
         if ((temp >= brkpoints[count].lowaddr)
                   && (temp <= brkpoints[count].highaddr))
           if (--brkcounts[count] == 0)
              memok = FALSE;
      }
   *seg += memmap[aseg];
   }

void checkwaddress1(seg1,ofs1,seg2,ofs2)
   unsigned int *seg1, *ofs1, *seg2, *ofs2;
   {
   UNSIGNED_LONG temp,temp2;
   unsigned int t1,t2,aseg1,aseg2,count;

   aseg1 = (t1 = *seg1 + (*ofs1 >> 4)) >> 12;
   aseg2 = (t2 = *seg2 + (*ofs2 >> 4)) >> 12;
   if (numbreaks)
      {
      temp = (((UNSIGNED_LONG)t1) << 4) + (*ofs1 & 0x000f);
      temp2 = (((UNSIGNED_LONG)t2) << 4) + (*ofs2 & 0x000f);
      for (count = 1; count <= numbreaks; count++)
         {
         if ((temp >= brkpoints[count].lowaddr)
                   && (temp <= brkpoints[count].highaddr))
           if (--brkcounts[count] == 0)
              memok = FALSE;
         if ((temp2 >= brkpoints[count].lowaddr)
                           && (temp2 <= brkpoints[count].highaddr))
            if (--brkcounts[count] == 0)
               memok = FALSE;
         }
      }
   *seg1 += memmap[aseg1];
   *seg2 += memmap[aseg2];
   }
