/*
   This file is copyright (c) 1986 by William Couture.  All rights reserved.
*/


#include <compiler.h>
#include <vim.h>
#include <regstruc.h>
#include <opdefs.h>

#define FLAG_MASK_8088 0xF000
#define BIT_3_ON '\010'   /* bit 3 */
#define BIT_3_OFF '\367'  /* bit 3 */
#define BIT_0_OFF  '\376'   /* bit 0 */
#define BIT_0_ON  '\001'  /* bit 0 */
/*
   I do not have the time to totally comment all functions in this file.  For
   more information, check out an 8088/8086 reference manual.  For information
   on how the flags are done, see the file asmfns.c.

   There are some comments in here, however, so do not overlook this file.
*/

/*
   NOTE:  The usual check for a BYTE/WORD instruction is:
      if (instruction.flags & BYTE)
   The lack of a BYTE flag is taken to mean a word flag.
*/

extern union regunion regs;

extern struct optable instruction;

extern struct tofrom dest,source;

extern unsigned int prefix;
extern unsigned int memok;
extern unsigned int errmessage;
extern unsigned int wait_status;
extern unsigned int segoffset;
extern unsigned int interruptcount;
extern unsigned int hwintflag;
extern unsigned int hwintstack[8];
extern unsigned char port60queue[64];
extern unsigned int port60ptr;
extern unsigned int hw8259on;
extern unsigned int eoiwait;
extern unsigned int pclone;
extern unsigned int dmamap[16];
extern unsigned int breakonio;
extern unsigned int videobase;
extern unsigned int attribflipflop;
extern unsigned int start_ip;
extern unsigned int fake_retrace;
extern unsigned int retr_flag;
extern unsigned int port64hold;
extern unsigned int aadaam_val;

extern int mod,reg,rm;

extern unsigned char byte_one;  /* from VIM.C */

extern void (*checkaddress)();  /* need a pointer to checkaddress */
extern void (*checkwaddress)();  /* and to checkwaddress */

unsigned int fetchbyte(),fetchword(),fromstack();

void db()  /* A routine to stick a pointer to.  Will never be used. */
   {
   }

void aaa()
   {
   if ((regs.r.flag & AUXCARRY) || ((regs.r.ax.byte.al & 0x0f) > 9))
      {
      regs.r.ax.byte.al += (char) 6;
      regs.r.ax.byte.ah++;
      regs.r.flag |= AUXCARRY;
      }
   regs.r.ax.byte.al &= 0x0f;
   if (regs.r.flag & AUXCARRY)
      regs.r.flag |= CARRY;
   else
      regs.r.flag &= CARRYOFF;
   }

void aad()
   {
   dest.oldvalue = regs.r.ax.byte.al;
   regs.r.ax.byte.al += regs.r.ax.byte.ah * aadaam_val;
   regs.r.ax.byte.ah = 0;
   dest.value = regs.r.ax.byte.al;
   flags(BYTE);
   }

void aam()
   {
   unsigned int newcs,newip,holdsp;

   if (!aadaam_val)  /* divide by zero trap */
      {
      holdsp = regs.r.sp;
      newip = fetchword(0,0);
      newcs = fetchword(0,2);
      tostack(regs.r.flag | FLAG_MASK_8088);
      tostack(regs.r.cs);
      tostack(regs.r.ip);
      if (memok)
         {
         regs.r.flag &= INTERRUPTOFF;
         regs.r.flag &= TRAPOFF;
         regs.r.cs = newcs;
         regs.r.ip = newip;
         }
      else
         regs.r.sp = holdsp;
      }
   else
      {
      dest.oldvalue = regs.r.ax.byte.al;
      regs.r.ax.byte.ah = regs.r.ax.byte.al / aadaam_val;
      regs.r.ax.byte.al %= aadaam_val;
      dest.value = regs.r.ax.byte.al;
      flags(BYTE);
      }
   }

void aas()
   {
   if ((regs.r.flag & AUXCARRY) || ((regs.r.ax.byte.al & 0x0f) > 9))
      {
      regs.r.ax.byte.al -= 6;
      regs.r.ax.byte.ah--;
      regs.r.flag |= AUXCARRY;
      }
   regs.r.ax.byte.al &= 0x0f;
   if (regs.r.flag & AUXCARRY)
      regs.r.flag |= CARRY;
   else
      regs.r.flag &= CARRYOFF;
   }

void adc()
   {
   UNSIGNED_LONG carryvalue;

   if (regs.r.flag & CARRY)
      carryvalue = 1;
   else
      carryvalue = 0;
   dest.value += source.value + carryvalue;
   if (instruction.flags & BYTE)
      {
      ovcarry((unsigned int)source.value,(unsigned int)carryvalue,BYTE);
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      ovcarry((unsigned int)source.value,(unsigned int)carryvalue,WORD);
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void add()
   {
   dest.value += source.value;
   if (instruction.flags & BYTE)
      {
      ovcarry((unsigned int)source.value,0,BYTE);
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      ovcarry((unsigned int)source.value,0,WORD);
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void and()
   {
   dest.value &= source.value;
   if (instruction.flags & BYTE)
      {
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   regs.r.flag &= CARRYOFF;
   regs.r.flag &= OVERFLOWOFF;
   }

void bound()
   {
   int lower,upper,regvalue;
   unsigned int newcs,newip,holdsp;

   lower = fetchword(source.seg,source.ofs);
   upper = fetchword(source.seg,source.ofs+2);
   if (memok)
      {
      regvalue = (unsigned int)dest.value;
      if ((regvalue < lower) || (regvalue > upper))
         {                            /* out of bounds, do an INT 5 */
         holdsp = regs.r.sp;
         newip = fetchword(0,0x14);
         newcs = fetchword(0,0x16);
         tostack(regs.r.flag | FLAG_MASK_8088);
         tostack(regs.r.cs);
         tostack(start_ip);  /* start of THIS instruction! */
         if (memok)
            {
            regs.r.flag &= INTERRUPTOFF;
            regs.r.flag &= TRAPOFF;
            regs.r.ip = newip;
            regs.r.cs = newcs;
            }
         else
            regs.r.sp = holdsp;
         }
      }
   }

void call()
   {
   tostack(regs.r.ip);
   if (memok)  /* did the stack hit restricted space? */
      regs.r.ip = (unsigned int)source.value;
   }

void callf()
   {
   unsigned int holdsp;

   holdsp = regs.r.sp;
   tostack(regs.r.cs);
   tostack(regs.r.ip);
   if (memok)  /* did the stack hit restricted space? */
      {
      regs.r.ip = (unsigned int)(source.value & 0xffffL);
      regs.r.cs = (unsigned int)((source.value >> 16) & 0xffffL);
      }
   else
      regs.r.sp = holdsp;
   }

void cbw()
   {
   if (regs.r.ax.byte.al & SIGNEDB)
      regs.r.ax.byte.ah = 0xff;
   else
      regs.r.ax.byte.ah = 0;
   }

void clc()
   {
   regs.r.flag &= CARRYOFF;
   }

void cld()
   {
   regs.r.flag &= DIRECTIONOFF;
   }

void cli()
   {
   regs.r.flag &= INTERRUPTOFF;
   }

void cmc()
   {
   regs.r.flag ^= CARRY;
   }

void cmp()
   {
   dest.value -= source.value;
   if (instruction.flags & BYTE)
      {
      ovcarry(((unsigned int)~source.value),1,BYTE);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(BYTE);
      }
   else
      {
      ovcarry(((unsigned int)~source.value),1,WORD);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(WORD);
      }
   }

void cmpsb()  /* In this string primitive, as in ALL string primitives,
                 there are essentially 2 or 3 versions of the routine, based
                 on a REPZ, REPNZ, and no repeat prefix conditions.
                 If the instruction merely checks for REP and does not care
                 about the zero flag (FORGETZF set), then there are 2
                 versions.
                 This was easier to write, though not as space efficient.
               */
   {
   unsigned int sourceseg,inc;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         }
   else
      sourceseg = regs.r.ds;
   if (regs.r.flag & DIRECTION)
      inc = -1;
   else
      inc = 1;
   if (prefix & REPZERO)
      {
      prefix |= ISZFON;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
         dest.oldvalue = dest.value;
         source.value = (UNSIGNED_LONG)fetchbyte(regs.r.es,regs.r.di);
         dest.value -= source.value;
         if (memok)
            {
            regs.r.si += inc;
            regs.r.di += inc;
            ovcarry(((unsigned int)~source.value),1,BYTE);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(BYTE);
            }
         }
      else
         prefix = 0;
      }
   else
      if (prefix & REPNOTZERO)
         {
         prefix |= ISZFOFF;
         if (regs.r.cx.word)
            {
            dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
            dest.oldvalue = dest.value;
            source.value = (UNSIGNED_LONG)fetchbyte(regs.r.es,regs.r.di);
            dest.value -= source.value;
            if (memok)
               {
               regs.r.si += inc;
               regs.r.di += inc;
               ovcarry(((unsigned int)~source.value),1,BYTE);
               regs.r.flag ^= CARRY;
               regs.r.flag ^= AUXCARRY;
               flags(BYTE);
               }
            }
         else
            prefix = 0;
         }
      else
         {
         dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
         dest.oldvalue = dest.value;
         source.value = (UNSIGNED_LONG)fetchbyte(regs.r.es,regs.r.di);
         dest.value -= source.value;
         if (memok)
            {
            regs.r.si += inc;
            regs.r.di += inc;
            ovcarry(((unsigned int)~source.value),1,BYTE);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(BYTE);
            }
         }
   }

void cmpsw()
   {
   unsigned int sourceseg,inc;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         }
   else
      sourceseg = regs.r.ds;
   if (regs.r.flag & DIRECTION)
      inc = -2;
   else
      inc = 2;
   if (prefix & REPZERO)
      {
      prefix |= ISZFON;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
         dest.oldvalue = dest.value;
         source.value = (UNSIGNED_LONG)fetchword(regs.r.es,regs.r.di);
         dest.value -= source.value;
         if (memok)
            {
            regs.r.si += inc;
            regs.r.di += inc;
            ovcarry(((unsigned int)~source.value),1,WORD);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(WORD);
            }
         }
      else
         prefix = 0;
      }
   else
      if (prefix & REPNOTZERO)
         {
         prefix |= ISZFOFF;
         if (regs.r.cx.word)
            {
            dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
            dest.oldvalue = dest.value;
            source.value = (UNSIGNED_LONG)fetchword(regs.r.es,regs.r.di);
            dest.value -= source.value;
            if (memok)
               {
               regs.r.si += inc;
               regs.r.di += inc;
               ovcarry(((unsigned int)~source.value),1,WORD);
               regs.r.flag ^= CARRY;
               regs.r.flag ^= AUXCARRY;
               flags(WORD);
               }
            }
         else
            prefix = 0;
         }
      else
         {
         dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
         dest.oldvalue = dest.value;
         source.value = (UNSIGNED_LONG)fetchword(regs.r.es,regs.r.di);
         dest.value -= source.value;
         if (memok)
            {
            regs.r.si += inc;
            regs.r.di += inc;
            ovcarry(((unsigned int)~source.value),1,WORD);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(WORD);
            }
         }
   }

void cwd()
   {
   if (regs.r.ax.word & SIGNEDW)
      regs.r.dx.word = 0xffff;
   else
      regs.r.dx.word = 0;
   }

void daa()
   {
   dest.oldvalue = regs.r.ax.byte.al;
   if ((regs.r.flag & AUXCARRY) || ((regs.r.ax.byte.al & 0x0f) > 9))
      {
      regs.r.ax.byte.al += 6;
      regs.r.flag |= AUXCARRY;
      }
   if ((regs.r.flag & CARRY) || ((regs.r.ax.byte.al & 0xf0) > 0x90))
      {
      regs.r.ax.byte.al += 0x60;
      regs.r.flag |= CARRY;
      }
   dest.value = regs.r.ax.byte.al;
   flags(BYTE);
   }

void das()
   {
   dest.oldvalue = regs.r.ax.byte.al;
   if ((regs.r.flag & AUXCARRY) || ((regs.r.ax.byte.al & 0x0f) > 9))
      {
      regs.r.ax.byte.al -= 6;
      regs.r.flag |= AUXCARRY;
      }
   if ((regs.r.flag & CARRY) || ((regs.r.ax.byte.al & 0xf0) > 0x90))
      {
      regs.r.ax.byte.al -= 0x60;
      regs.r.flag |= CARRY;
      }
   dest.value = regs.r.ax.byte.al;
   flags(BYTE);
   }

void dec()
   {
   unsigned int carrysave;

   dest.value--;
   if (instruction.flags & BYTE)
      {
      carrysave = regs.r.flag & CARRY;
      ovcarry(0xffff,0,BYTE);
      flags(BYTE);
      regs.r.flag ^= AUXCARRY;
      regs.r.flag &= CARRYOFF;
      regs.r.flag |= carrysave;
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      carrysave = regs.r.flag & CARRY;
      ovcarry(0xffff,0,WORD);
      flags(WORD);
      regs.r.flag ^= AUXCARRY;
      regs.r.flag &= CARRYOFF;
      regs.r.flag |= carrysave;
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void div()
   {
   UNSIGNED_LONG temp1,temp2,ul_div(),ul_mod();
   unsigned int newcs,newip,holdsp;

   temp1 = 0;
   temp2 = 0;
   if (instruction.flags & BYTE)
      {
      if (source.value)
         {
         temp1 = regs.r.ax.word / source.value;
         temp2 = regs.r.ax.word % source.value;
         }
      if ((!source.value) || (temp1 > 0xffL))  /* divide by zero trap */
         {
         holdsp = regs.r.sp;
         newip = fetchword(0,0);
         newcs = fetchword(0,2);
         tostack(regs.r.flag | FLAG_MASK_8088);
         tostack(regs.r.cs);
         tostack(regs.r.ip);
         if (memok)
            {
            regs.r.flag &= INTERRUPTOFF;
            regs.r.flag &= TRAPOFF;
            regs.r.cs = newcs;
            regs.r.ip = newip;
            }
         else
            regs.r.sp = holdsp;
         }
      else
         {
         regs.r.ax.byte.al = temp1;
         regs.r.ax.byte.ah = temp2;
         }
      }
   else
      {
      if (source.value)
         {
         temp1 = regs.r.dx.word;
         temp1 <<= 16;
         temp1 += regs.r.ax.word;
         temp2 = temp1;
         temp1 = ul_div(temp1,source.value);
         temp2 = ul_mod(temp2,source.value);
         }
      if ((!source.value) || (temp1 > 0xffffL))  /* divide by 0 trap */
         {
         holdsp = regs.r.sp;
         newip = fetchword(0,0);
         newcs = fetchword(0,2);
         tostack(regs.r.flag | FLAG_MASK_8088);
         tostack(regs.r.cs);
         tostack(regs.r.ip);
         if (memok)
            {
            regs.r.flag &= INTERRUPTOFF;
            regs.r.flag &= TRAPOFF;
            regs.r.cs = newcs;
            regs.r.ip = newip;
            }
         else
            regs.r.sp = holdsp;
         }
      else
         {
         regs.r.ax.word = temp1;
         regs.r.dx.word = temp2;
         }
      }
   }

void enter()
   {
   unsigned int holdsp,tempbp,tempss,count,frame;

   holdsp = regs.r.sp;
   tempbp = regs.r.bp;
   tostack(regs.r.bp);
   if (memok)
      {
      frame = regs.r.sp;
      count = (unsigned char)source.value;
      if (count)
         {
         while (--count)
            {
            tempbp -= 2;
            tostack(fetchword(regs.r.ss,tempbp));
            }
         tostack(frame);
         }
      if (memok)
         {
         regs.r.bp = frame;
         regs.r.sp -= (unsigned int)dest.value;
         }
      else
         regs.r.sp = holdsp;
      }
   }

void esc()  /* coprocessors not supported */
   {
   }

void hlt()
   {
   errmessage = HALTERR;
   regs.r.ip--;  /* force back to the HLT instruction */
   }

void idiv()
   {
   long temp1,temp2;
   unsigned int newcs,newip,holdsp;

   temp1 = 0;
   temp2 = 0;
   if (instruction.flags & BYTE)
      {
      if (source.value)
         {
         if (source.value & 0x80L)
            source.value |= 0xff00L;  /* sign extend to get proper signed
                               result.  May have problems on your compiler. */
         if (regs.r.ax.word & SIGNEDW)
            temp1 = 0xffff0000L | regs.r.ax.word;
         else
            temp1 = regs.r.ax.word;
         temp2 = temp1;
         temp1 /= source.value;
         temp2 %= source.value;
         }
      if ((!source.value) || (temp1 > 0x7fL))  /* divide by 0 trap */
         {
         holdsp = regs.r.sp;
         newip = fetchword(0,0);
         newcs = fetchword(0,2);
         tostack(regs.r.flag | FLAG_MASK_8088);
         tostack(regs.r.cs);
         tostack(regs.r.ip);
         if (memok)
            {
            regs.r.flag &= INTERRUPTOFF;
            regs.r.flag &= TRAPOFF;
            regs.r.cs = newcs;
            regs.r.ip = newip;
            }
         else
            regs.r.sp = holdsp;
         }
      else
         {
         regs.r.ax.byte.al = temp1;
         regs.r.ax.byte.ah = temp2;
         }
      }
   else
      {
      if (source.value)
         {
         if (source.value & 0x8000L)
            source.value |= 0xffff0000L;  /* Sign extension again.
                                             Watch it! */
         temp1 = regs.r.dx.word;
         temp1 <<= 16;
         temp1 += regs.r.ax.word;
         temp2 = temp1;
         temp1 /= source.value;
         temp2 %= source.value;
         }
      if ((!source.value) || (temp1 > 0x7fffL))  /* divide by 0 trap */
         {
         holdsp = regs.r.sp;
         newip = fetchword(0,0);
         newcs = fetchword(0,2);
         tostack(regs.r.flag | FLAG_MASK_8088);
         tostack(regs.r.cs);
         tostack(regs.r.ip);
         if (memok)
            {
            regs.r.flag &= INTERRUPTOFF;
            regs.r.flag &= TRAPOFF;
            regs.r.cs = newcs;
            regs.r.ip = newip;
            }
         else
            regs.r.sp = holdsp;
         }
      else
         {
         regs.r.ax.word = temp1;
         regs.r.dx.word = temp2;
         }
      }
   }

void imul()
   {
   long temp;

   if ((byte_one & '\375') == '\151')  /* if IMUL immediate */
      {
      regs.wordarray[reg] = (int)(((int)source.value) * ((int)dest.value));
      }
   else
      {
      if (instruction.flags & BYTE)
         {
         if (source.value & 0x80L)
            source.value |= 0xff00L;
         if (regs.r.ax.byte.al & SIGNEDB)
            temp = 0xff00L | regs.r.ax.byte.al;
         else
            temp = regs.r.ax.byte.al;
         temp *= source.value;
         regs.r.ax.byte.al = temp & 0xffL;
         regs.r.ax.byte.ah = (temp >> 8) & 0xffL;
         if ((regs.r.ax.byte.ah == 0xff) || (regs.r.ax.byte.ah == 0))
            {
            regs.r.flag &= OVERFLOWOFF;
            regs.r.flag &= CARRYOFF;
            }
         else
            {
            regs.r.flag |= OVERFLOW;
            regs.r.flag |= CARRY;
            }
         }
      else
         {
         if (source.value & 0x8000L)
            source.value |= 0xffff0000L;
         if (regs.r.ax.word & SIGNEDW)
            temp = 0xffff0000L | regs.r.ax.word;
         else
            temp = regs.r.ax.word;
         temp *= source.value;
         regs.r.ax.word = temp & 0xffffL;
         regs.r.dx.word = (temp >> 16) & 0xffffL;
         if ((regs.r.dx.word == 0xffff) || (regs.r.dx.word == 0))
            {
            regs.r.flag &= OVERFLOWOFF;
            regs.r.flag &= CARRYOFF;
            }
         else
            {
            regs.r.flag |= OVERFLOW;
            regs.r.flag |= CARRY;
            }
         }
      }
   }

void in()
   {
   if (instruction.source == SDX)  /* handle the special case of
             IN AL,DX, where both BYTE and WORD flags need to be set */
      source.value = (UNSIGNED_LONG)regs.r.dx.word;
   if (breakonio)
      {
      if (instruction.flags & BYTE)
         checkport(BYTE,(unsigned int)source.value);
      else
         checkport(WORD,(unsigned int)source.value);
      }
   if (memok)
      {
      if (pclone)  /* let the request pass to the hardware */
         {
         if (instruction.flags & BYTE)
              /* Note that special ports are ONLY checked for on byte ports! */
            {
            if (source.value == ((UNSIGNED_LONG)videobase+6))
               attribflipflop = 0;
            if (source.value == 0x60L)  /* the keyboard port has to be saved, as
                        the BIOS routine clears the port */
               {
               regs.r.ax.byte.al = port60queue[0];
               port64hold = INBPORT(0x64);
               }
            else
               if (source.value == 0x64L)
                  {
                  regs.r.ax.byte.al = port64hold;
                  port64hold = INBPORT(0x64);
                  }
               else
                  regs.r.ax.byte.al = INBPORT((unsigned int)source.value);
            if ((fake_retrace) &&
                  ((((unsigned int)source.value) == 0x3ba) ||
                   (((unsigned int)source.value) == 0x3da)))
               if (retr_flag)  /* retrace mode */
                  regs.r.ax.byte.al = (regs.r.ax.byte.al | BIT_3_ON)
                                           & BIT_0_OFF;
               else  /* video mode */
                  regs.r.ax.byte.al = (regs.r.ax.byte.al & BIT_3_OFF)
                                           | BIT_0_ON;
            }
         else
            {
            if ((source.value == ((UNSIGNED_LONG)videobase+5))
                       || (source.value == ((UNSIGNED_LONG)videobase+6)))
               attribflipflop = 0;
            regs.r.ax.word = INWPORT((unsigned int)source.value);
            }
         }
      else
         if (instruction.flags & BYTE)
            regs.r.ax.byte.al = 0xff;
         else
            regs.r.ax.word = 0xffff;
      }
   }

void inc()
   {
   unsigned int carrysave;

   dest.value++;
   if (instruction.flags & BYTE)
      {
      carrysave = regs.r.flag & CARRY;
      ovcarry(1,0,BYTE);
      flags(BYTE);
      regs.r.flag &= CARRYOFF;
      regs.r.flag |= carrysave;
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      carrysave = regs.r.flag & CARRY;
      ovcarry(1,0,WORD);
      flags(WORD);
      regs.r.flag &= CARRYOFF;
      regs.r.flag |= carrysave;
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void insb()
   {
   unsigned int inc,hold,seg,ofs;

   if (breakonio)
      checkport(BYTE,regs.r.dx.word);
   if (memok)
      {
      if (regs.r.flag & DIRECTION)
         inc = -1;
      else
         inc = 1;
      if (prefix & REP)
         {
         prefix |= FORGETZF;
         if (regs.r.cx.word)
            {
            seg = regs.r.es;
            ofs = regs.r.di;
            (*checkaddress)(&seg,&ofs);
            if (pclone)
               {
     /*
        NOTE:  The destination address MUST be checked BEFORE the port fetch
        is performed!  The reason for this is that repeated fetches from the
        same port obviously tell the hardware to return the next value in the
        buffer.  If a value is fetched and then the destination address is
        found to be a trap point, the value fetched will be discareded.  When
        execution starts again, the fetch will be re-performed, resulting in
        the wrong value for this and all subsequent bytes!
   */
               if (memok)
                  {
                  if (regs.r.dx.word == 0x60)  /* the keyboard port has to be
                                 saved, as the BIOS routine clears the port */
                     {
                     hold = port60queue[0];
                     port64hold = INBPORT(0x64);
                     }
                  else
                     if (regs.r.dx.word == 0x64)
                        {
                        hold = port64hold;
                        port64hold = INBPORT(0x64);
                        }
                     else
                        hold = INBPORT(regs.r.dx.word);
                  }
               }
            else
               hold = 0xff;  /* empty port, return all ones */
/*            putbyte(regs.r.es,regs.r.di,hold); */
            if (memok)
               {
               POKE(hold,seg,ofs);  /* put into alread-checked memory */
               if (regs.r.dx.word == videobase + 6)
                  attribflipflop = 0;
               regs.r.di += inc;
               }
            }
         else
            prefix = 0;
         }
      else
         {
         seg = regs.r.es;
         ofs = regs.r.di;
         (*checkaddress)(&seg,&ofs);  /* see comments above */
         if (pclone)
            {
            if (memok)
               {
               if (regs.r.dx.word == 0x60)
                  {
                  hold = port60queue[0];
                  port64hold = INBPORT(0x64);
                  }
               else
                  if (regs.r.dx.word == 0x64)
                     {
                     hold = port64hold;
                     port64hold = INBPORT(0x64);
                     }
                  else
                     hold = INBPORT(regs.r.dx.word);
               }
            }
         else
            hold = 0xff;
/*         putbyte(regs.r.es,regs.r.di,hold); */
         if (memok)
            {
            POKE(hold,seg,ofs);  /* put into alread-checked memory */
            if (regs.r.dx.word == videobase + 6)
               attribflipflop = 0;
            regs.r.di += inc;
            }
         }
      }
   }

void insw()
   {
   unsigned int inc,hold,seg1,ofs1,seg2,ofs2;

   if (breakonio)
      checkport(WORD,regs.r.dx.word);
   if (memok)
      {
      if (regs.r.flag & DIRECTION)
         inc = -2;
      else
         inc = 2;
      if (prefix & REP)
         {
         prefix |= FORGETZF;
         if (regs.r.cx.word)
            {
            seg1 =
            seg2 = regs.r.es;
            ofs1 = regs.r.di;
            ofs2 = regs.r.di+1;
            (*checkwaddress)(&seg1,&ofs1,&seg2,&ofs2);  /* see comments above in INSB() */
/*            putword(regs.r.es,regs.r.di,hold); */
            if (memok)
               {
               if (pclone)
                  hold = INWPORT(regs.r.dx.word);
               else
                  hold = 0xffff;
               POKE(hold,seg1,ofs1);  /* low byte */
               POKE(hold>>8,seg2,ofs2);  /* high byte */
               if ((regs.r.dx.word == videobase+6) || (regs.r.dx.word == videobase+5))
                  attribflipflop = 0;
               regs.r.di += inc;
               }
            }
         else
            prefix = 0;
         }
      else
         {
         seg1 =
         seg2 = regs.r.es;
         ofs1 = regs.r.di;
         ofs2 = regs.r.di+1;
         (*checkwaddress)(&seg1,&ofs1,&seg2,&ofs2);  /* see comments above in INSB() */
/*         putword(regs.r.es,regs.r.di,hold); */
         if (memok)
            {
            if (pclone)
               hold = INWPORT(regs.r.dx.word);
            else
               hold = 0xffff;
            POKE(hold,seg1,ofs1);  /* low byte */
            POKE(hold>>8,seg2,ofs2);  /* high byte */
            if ((regs.r.dx.word == videobase+6) || (regs.r.dx.word == videobase+5))
               attribflipflop = 0;
            regs.r.di += inc;
            }
         }
      }
   }

void intr()  /* The INT instruction, called intr() to distinguish it from
                int variables. */
   {
   unsigned int newcs, newip, holdsp;

   holdsp = regs.r.sp;
   if (instruction.flags & INT3)  /* if hardwired to INT 3 */
      {
      newip = fetchword(0,12);
      newcs = fetchword(0,14);
      tostack(regs.r.flag | FLAG_MASK_8088);
      tostack(regs.r.cs);
      tostack(regs.r.ip);
      }
   else
      {
      newip = fetchword(0,((unsigned int) source.value*4));
      newcs = fetchword(0,((unsigned int) source.value*4+2));
      tostack(regs.r.flag | FLAG_MASK_8088);
      tostack(regs.r.cs);
      tostack(regs.r.ip);
      }
   if (memok)
      {
      regs.r.flag &= INTERRUPTOFF;
      regs.r.flag &= TRAPOFF;
      regs.r.ip = newip;
      regs.r.cs = newcs;
      }
   else
      regs.r.sp = holdsp;
   }

void into()
   {
   unsigned int newip, newcs, holdsp;

   holdsp = regs.r.sp;
   if (regs.r.flag & OVERFLOW)
      {
      newip = fetchword(0,16);
      newcs = fetchword(0,18);
      tostack(regs.r.flag | FLAG_MASK_8088);
      tostack(regs.r.cs);
      tostack(regs.r.ip);
      if (memok)
         {
         regs.r.flag &= INTERRUPTOFF;
         regs.r.flag &= TRAPOFF;
         regs.r.ip = newip;
         regs.r.cs = newcs;
         }
      else
         regs.r.sp = holdsp;
      }
   }

void iret()
   {
   unsigned int newcs, newip, newflag, holdsp;

   holdsp = regs.r.sp;
   newip = fromstack();
   newcs = fromstack();
   newflag = fromstack();
   if (memok)
      {
      regs.r.flag = newflag;
      regs.r.cs = newcs;
      regs.r.ip = newip;
      if (regs.r.flag & TRAP)  /* if trap was just set */
         interruptcount = 1;  /* let the next instruction execute before
                                 trapping */
      }
   else
      regs.r.sp = holdsp;
   }

void ja()
   {
   if ( ((regs.r.flag & CARRY) == 0) && ((regs.r.flag & ZERO) == 0) )
      regs.r.ip = (unsigned int)source.value;
   }

void jae()
   {
   if ((regs.r.flag & CARRY) == 0)
      regs.r.ip = (unsigned int)source.value;
   }

void jcxz()
   {
   if (regs.r.cx.word == 0)
      regs.r.ip = (unsigned int)source.value;
   }

void jg()
   {
   if (!(regs.r.flag & ZERO))
      if ( ((regs.r.flag & SIGN) && (regs.r.flag & OVERFLOW))
       || (((regs.r.flag & SIGN) == 0) && ((regs.r.flag & OVERFLOW) == 0)) )
         regs.r.ip = (unsigned int)source.value;
   }

void jge()
   {
   if ( ((regs.r.flag & SIGN) && (regs.r.flag & OVERFLOW))
        || (((regs.r.flag & SIGN) == 0) && ((regs.r.flag & OVERFLOW) == 0)) )
      regs.r.ip = (unsigned int)source.value;
   }

void jmp()
   {
   regs.r.ip = (unsigned int)source.value;
   }

void jmpf()
   {
   regs.r.ip = (unsigned int)(source.value & 0xffffL);
   regs.r.cs = (unsigned int)((source.value >> 16) & 0xffffL);
   }

void jna()
   {
   if ((regs.r.flag & CARRY) || (regs.r.flag & ZERO))
      regs.r.ip = (unsigned int)source.value;
   }

void jnae()
   {
   if (regs.r.flag & CARRY)
      regs.r.ip = (unsigned int)source.value;
   }

void jng()
   {
   if ((regs.r.flag & ZERO)
           || (((regs.r.flag & SIGN) == 0) && (regs.r.flag & OVERFLOW))
           || ((regs.r.flag & SIGN) && ((regs.r.flag & OVERFLOW) == 0)) )
      regs.r.ip = (unsigned int)source.value;
   }

void jnge()
   {
   if ( ((regs.r.flag & SIGN) && ((regs.r.flag & OVERFLOW) == 0))
        || (((regs.r.flag & SIGN) == 0) && (regs.r.flag & OVERFLOW)) )
      regs.r.ip = (unsigned int)source.value;
   }

void jno()
   {
   if ((regs.r.flag & OVERFLOW) == 0)
      regs.r.ip = (unsigned int)source.value;
   }

void jns()
   {
   if ((regs.r.flag & SIGN) == 0)
      regs.r.ip = (unsigned int)source.value;
   }

void jnz()
   {
   if ((regs.r.flag & ZERO) == 0)
      regs.r.ip = (unsigned int)source.value;
   }

void jo()
   {
   if (regs.r.flag & OVERFLOW)
      regs.r.ip = (unsigned int)source.value;
   }

void jpe()
   {
   if (regs.r.flag & PARITY)
      regs.r.ip = (unsigned int)source.value;
   }

void jpo()
   {
   if ((regs.r.flag & PARITY) == 0)
      regs.r.ip = (unsigned int)source.value;
   }

void js()
   {
   if (regs.r.flag & SIGN)
      regs.r.ip = (unsigned int)source.value;
   }

void jz()
   {
   if (regs.r.flag & ZERO)
      regs.r.ip = (unsigned int)source.value;
   }

void lahf()
   {
   regs.r.ax.byte.ah = regs.r.flag & 0xff;
   }

void lds()
   {
   unsigned int tempreg,tempseg;

   tempreg = source.value;
   tempseg = fetchword(source.seg,source.ofs+2);
   if (memok)
      {
      regs.wordarray[dest.reg] = tempreg;
      regs.r.ds = tempseg;
      }
   }

void lea()
   {
   regs.wordarray[dest.reg] = source.ofs;
   }

void leave()
   {
   unsigned int holdsp,tempbp;

   holdsp = regs.r.sp;
   regs.r.sp = regs.r.bp;
   tempbp = fromstack();
   if (memok)
      regs.r.bp = tempbp;
   else
      regs.r.sp = holdsp;
   }

void les()
   {
   unsigned int tempreg,tempseg;

   tempreg = source.value;
   tempseg = fetchword(source.seg,source.ofs+2);
   if (memok)
      {
      regs.wordarray[dest.reg] = tempreg;
      regs.r.es = tempseg;
      }
   }

void lock()
   {
   }

void lodsb()
   {
   unsigned int sourceseg,inc;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         }
   else
      sourceseg = regs.r.ds;
   if (regs.r.flag & DIRECTION)
      inc = -1;
   else
      inc = 1;
   if (prefix & REP)
      {
      prefix |= FORGETZF;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
         if (memok)
            {
            regs.r.si += inc;
            regs.r.ax.byte.al = dest.value;
            }
         }
      else
         prefix = 0;
      }
   else
      {
      dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
      if (memok)
         {
         regs.r.si += inc;
         regs.r.ax.byte.al = dest.value;
         }
      }
   }

void lodsw()
   {
   unsigned int sourceseg,inc;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         }
   else
      sourceseg = regs.r.ds;
   if (regs.r.flag & DIRECTION)
      inc = -2;
   else
      inc = 2;
   if (prefix & REP)
      {
      prefix |= FORGETZF;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
         if (memok)
            {
            regs.r.si += inc;
            regs.r.ax.word = dest.value;
            }
         }
      else
         prefix = 0;
      }
   else
      {
      dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
      if (memok)
         {
         regs.r.si += inc;
         regs.r.ax.word = dest.value;
         }
      }
   }

void loop()
   {
   regs.r.cx.word--;
   if (regs.r.cx.word)
      regs.r.ip = (unsigned int)source.value;
   }

void loopnz()
   {
   regs.r.cx.word--;
   if ((regs.r.cx.word) && (regs.r.flag & ZERO))
      regs.r.ip = (unsigned int)source.value;
   }

void loopz()
   {
   regs.r.cx.word--;
   if ((regs.r.cx.word) && ((regs.r.flag & ZERO) == 0))
      regs.r.ip = (unsigned int)source.value;
   }

void mov()
   {
   if (instruction.flags & BYTE)
      {
      if (dest.where & REG)
         regs.bytearray[dest.reg] = source.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)source.value));
      }
   else
      {
      if (dest.where & REG)
         {
         regs.wordarray[dest.reg] = source.value;
         if (dest.where & SEGREG)
            interruptcount = 1;
         }
      else
         putword(dest.seg,dest.ofs,((unsigned int)source.value));
      }
   }

void movsb()
   {
   unsigned int sourceseg,inc;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         }
   else
      sourceseg = regs.r.ds;
   if (regs.r.flag & DIRECTION)
      inc = -1;
   else
      inc = 1;
   if (prefix & REP)
      {
      prefix |= FORGETZF;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
         if (memok)
            {
            putbyte(regs.r.es,regs.r.di,((unsigned int)dest.value));
            if (memok)
               {
               regs.r.si += inc;
               regs.r.di += inc;
               }
            }
         }
      else
         prefix = 0;
      }
   else
      {
      dest.value = (UNSIGNED_LONG)fetchbyte(sourceseg,regs.r.si);
      if (memok)
         {
         putbyte(regs.r.es,regs.r.di,((unsigned int)dest.value));
         if (memok)
            {
            regs.r.si += inc;
            regs.r.di += inc;
            }
         }
      }
   }

void movsw()
   {
   unsigned int sourceseg,inc;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         }
   else
      sourceseg = regs.r.ds;
   if (regs.r.flag & DIRECTION)
      inc = -2;
   else
      inc = 2;
   if (prefix & REP)
      {
      prefix |= FORGETZF;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
         if (memok)
            {
            putword(regs.r.es,regs.r.di,((unsigned int)dest.value));
            if (memok)
               {
               regs.r.si += inc;
               regs.r.di += inc;
               }
            }
         }
      else
         prefix = 0;
      }
   else
      {
      dest.value = (UNSIGNED_LONG)fetchword(sourceseg,regs.r.si);
      if (memok)
         {
         putword(regs.r.es,regs.r.di,((unsigned int)dest.value));
         if (memok)
            {
            regs.r.si += inc;
            regs.r.di += inc;
            }
         }
      }
   }

void mul()
   {
   long temp;

   if (instruction.flags & BYTE)
      {
      temp = regs.r.ax.byte.al * source.value;
      regs.r.ax.byte.al = temp & 0xffL;
      regs.r.ax.byte.ah = (temp >> 8) & 0xffL;
      flags(BYTE);
      if (regs.r.ax.byte.ah == 0)
         {
         regs.r.flag &= OVERFLOWOFF;
         regs.r.flag &= CARRYOFF;
         }
      else
         {
         regs.r.flag |= OVERFLOW;
         regs.r.flag |= CARRY;
         }
      }
   else
      {
      temp = regs.r.ax.word * source.value;
      regs.r.ax.word = temp & 0xffffL;
      regs.r.dx.word = (temp >> 16) & 0xffffL;
      flags(WORD);
      if (regs.r.dx.word == 0)
         {
         regs.r.flag &= OVERFLOWOFF;
         regs.r.flag &= CARRYOFF;
         }
      else
         {
         regs.r.flag |= OVERFLOW;
         regs.r.flag |= CARRY;
         }
      }
   }

void neg()
   {
   dest.value = -dest.value;
   dest.oldvalue ^= 0xffffffffL;
   if (instruction.flags & BYTE)
      {
      ovcarry(1,0,BYTE);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      ovcarry(1,0,WORD);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void nop()
   {
   }

void not()
   {
   dest.value = ~dest.value;
   if (instruction.flags & BYTE)
      {
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void or()
   {
   dest.value |= source.value;
   if (instruction.flags & BYTE)
      {
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   regs.r.flag &= CARRYOFF;
   regs.r.flag &= OVERFLOWOFF;
   }

void out()
   {
   if (instruction.dest == DDX)
      dest.value = (UNSIGNED_LONG)regs.r.dx.word;  /* handle the special
          case of OUT DX,AL  where both BYTE and WORD flags need to be set */
   if (breakonio)
      {
      if (instruction.flags & BYTE)
         checkport(BYTE,(unsigned int)dest.value);
      else
         checkport(WORD,(unsigned int)dest.value);
      }
   if (memok)
      {
      if (pclone)  /* If doing an XT clone */
         {
         if (instruction.flags & BYTE)  /* both BYTE and WORD are set for
                               OUT DX,AL, but BYTE is the one we want here */
            outmap( ((unsigned int)dest.value),regs.r.ax.byte.al,BYTE);
         else
            outmap( ((unsigned int)dest.value),regs.r.ax.word,WORD);
         }
   /* else toss it into the bit bucket */
      }
   }

void outsb()
   {
   unsigned int sourceseg,inc,hold;

   if (breakonio)
      checkport(BYTE,regs.r.dx.word);
   if (memok)
      {
      if (prefix & SEG)
         switch (prefix & SEG)
            {
            case ESSEG:
               sourceseg = regs.r.es;
               break;
            case CSSEG:
               sourceseg = regs.r.cs;
               break;
            case SSSEG:
               sourceseg = regs.r.ss;
               break;
            case DSSEG:
               sourceseg = regs.r.ds;
               break;
            }
      else
         sourceseg = regs.r.ds;
      if (regs.r.flag & DIRECTION)
         inc = -1;
      else
         inc = 1;
      if (prefix & REP)
         {
         prefix |= FORGETZF;
         if (regs.r.cx.word)
            {
            hold = fetchbyte(sourceseg,regs.r.si);
            if (memok)
               {
               regs.r.si += inc;
               if (pclone)  /* if PC-XT clone status enabled */
                  outmap(regs.r.dx.word,hold,BYTE);
      /* else toss it in the bit bucket */
               }
            }
         else
            prefix = 0;
         }
      else
         {
         hold = fetchbyte(sourceseg,regs.r.si);
         if (memok)
            {
            regs.r.si += inc;
            if (pclone)  /* if PC-XT clone enabled */
               outmap(regs.r.dx.word,hold,BYTE);
      /* else toss it in the bit bucket */
            }
         }
      }
   }
   
void outsw()
   {
   unsigned int sourceseg,inc,hold;

   if (breakonio)
      checkport(WORD,regs.r.dx.word);
   if (memok)
      {
      if (prefix & SEG)
         switch (prefix & SEG)
            {
            case ESSEG:
               sourceseg = regs.r.es;
               break;
            case CSSEG:
               sourceseg = regs.r.cs;
               break;
            case SSSEG:
               sourceseg = regs.r.ss;
               break;
            case DSSEG:
               sourceseg = regs.r.ds;
               break;
            }
      else
         sourceseg = regs.r.ds;
      if (regs.r.flag & DIRECTION)
         inc = -2;
      else
         inc = 2;
      if (prefix & REP)
         {
         prefix |= FORGETZF;
         if (regs.r.cx.word)
            {
            hold = fetchword(sourceseg,regs.r.si);
            if (memok)
               {
               regs.r.si += inc;
               if (pclone)  /* if PC-XT clone enabled */
                  outmap(regs.r.dx.word,hold,WORD);
      /* else toss it in the bit bucket */
               }
            }
         else
            prefix = 0;
         }
      else
         {
         hold = fetchword(sourceseg,regs.r.si);
         if (memok)
            {
            regs.r.si += inc;
            if (pclone)  /* if PC-XT clone enabled */
               outmap(regs.r.dx.word,hold,WORD);
      /* else toss it in the bit bucket */
            }
         }
      }
   }

void pop()
   {
   unsigned int temp;

   temp = fromstack();
   if (memok)
      {
      if (dest.where & REG)
         {
         regs.wordarray[dest.reg] = temp;
         if (dest.where & SEGREG)
            interruptcount = 1;
         }
      else
         putword(dest.seg,dest.ofs,temp);
      }
   }

void popa()  /* support some additional 80286 instructions */
   {
   unsigned int oldsp,holdax,holdbx,holdcx,holddx,holdsp,holdbp,holdsi,holddi;

   oldsp = regs.r.sp;  /* save sp for later */
   holddi = fromstack();  /* get all registers into temporary storage */
   holdsi = fromstack();
   holdbp = fromstack();
   holdsp = fromstack();
   holdbx = fromstack();
   holddx = fromstack();
   holdcx = fromstack();
   holdax = fromstack();
   if (memok)
      {
      regs.r.ax.word = holdax;  /* Set regisgers.  Note that SP takes care */
      regs.r.bx.word = holdbx;  /* of itself. */
      regs.r.cx.word = holdcx;
      regs.r.dx.word = holddx;
      regs.r.bp = holdbp;
      regs.r.si = holdsi;
      regs.r.di = holddi;
      }
   else
      regs.r.sp = oldsp;  /* if protected address was in the middle of that
                             batch, restore sp to previous value */
   }

void popf()
   {
   unsigned int temp;

   temp = fromstack();
   if (memok)
      {
      regs.r.flag = temp;
      if (regs.r.flag & TRAP)  /* if trap was just set */
         interruptcount = 1;  /* execute next instruction before trapping */
      }
   }

void push()
   {
   if ((instruction.source == SSP)
         || ((instruction.source == SM) && (mod == 3) && (rm == 4)))
      tostack((unsigned int)source.value-2);  /* push sp-2, not sp, to
                                                 emulate 8088 */
   else
      tostack((unsigned int)source.value);
   }

void pusha()
   {
   unsigned int count,tempss,tempsp;

/*   for (count = 1; count < 17; count++)
                      {* check memory addresses sp-1...sp-16 BEFORE pushing *}
      {
      tempsp = regs.r.sp - count;
      tempss = regs.r.ss;
      (*checkaddress)(&tempss,&tempsp);
      }
   if (memok)
      { */

      tempsp = regs.r.sp;
      tostack(regs.r.ax.word);
      tostack(regs.r.cx.word);
      tostack(regs.r.dx.word);
      tostack(regs.r.bx.word);
      tostack(tempsp);
      tostack(regs.r.bp);
      tostack(regs.r.si);
      tostack(regs.r.di);
      if (memok == FALSE)
         regs.r.sp = tempsp;

/*      } */
   }

void pushf()
   {
   tostack(regs.r.flag | FLAG_MASK_8088);
   }

void rcl()
   {
   UNSIGNED_LONG templ;
   unsigned int temp,carryvalue;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (regs.r.flag & CARRY)
      carryvalue = 1;
   else
      carryvalue = 0;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         templ = (dest.value << 1) + carryvalue;
         if (templ & 0x100L)
            {
            regs.r.flag |= CARRY;
            carryvalue = 1;
            if (dest.value & 0x40L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            regs.r.flag &= CARRYOFF;
            carryvalue = 0;
            if (dest.value & 0x40L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value = templ;
         }
      if (dest.where & REG)
         regs.bytearray[dest.reg] = templ;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)templ));
      }
   else
      {
      while (temp--)
         {
         templ = (dest.value << 1) + carryvalue;
         if (templ & 0x10000L)
            {
            regs.r.flag |= CARRY;
            carryvalue = 1;
            if (dest.value & 0x4000L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            regs.r.flag &= CARRYOFF;
            carryvalue = 0;
            if (dest.value & 0x4000L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value = templ;
         }
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }
   
void rcr()
   {
   unsigned int temp,carryvalue;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         if (regs.r.flag & CARRY)
            {
            carryvalue = SIGNEDB;
            if (dest.value & 0x80L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            carryvalue = 0;
            if (dest.value & 0x80L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         if (dest.value & 1)
            regs.r.flag |= CARRY;
         else
            regs.r.flag &= CARRYOFF;
         dest.value = ((dest.value >> 1) & 0x7fL) + carryvalue;
         }
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      while (temp--)
         {
         if (regs.r.flag & CARRY)
            {
            carryvalue = SIGNEDW;
            if (dest.value & 0x8000L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            carryvalue = 0;
            if (dest.value & 0x8000L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         if (dest.value & 1)
            regs.r.flag |= CARRY;
         else
            regs.r.flag &= CARRYOFF;
         dest.value = ((dest.value >> 1) & 0x7fffL) + carryvalue;
         }
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void repnz()
   {
   }

void repz()
   {
   }

void ret()
   {
   unsigned int tempip;

   tempip = fromstack();
   if (memok)
      {
      regs.r.ip = tempip;
      if (instruction.flags & ADJSTK)
         regs.r.sp += source.value;
      }
   }

void retf()
   {
   unsigned int tempip,tempcs,holdsp;

   holdsp = regs.r.sp;
   tempip = fromstack();
   tempcs = fromstack();
   if (memok)
      {
      regs.r.ip = tempip;
      regs.r.cs = tempcs;
      if (instruction.flags & ADJSTK)
         regs.r.sp += source.value;
      }
   else
      regs.r.sp = holdsp;
   }

void rol()
   {
   unsigned int temp,wrapflag;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         if (dest.value & 0x80L)
            {
            wrapflag = 1;
            regs.r.flag |= CARRY;
            if (dest.value & 0x40L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            wrapflag = 0;
            regs.r.flag &= CARRYOFF;
            if (dest.value & 0x40L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value = (dest.value << 1) + wrapflag;
         }
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      while (temp--)
         {
         if (dest.value & 0x8000L)
            {
            wrapflag = 1;
            regs.r.flag |= CARRY;
            if (dest.value & 0x4000L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            wrapflag = 0;
            regs.r.flag &= CARRYOFF;
            if (dest.value & 0x4000L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value = (dest.value << 1) + wrapflag;
         }
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void ror()
   {
   unsigned int temp,wrapflag;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         if (dest.value & 1)
            {
            wrapflag = SIGNEDB;
            regs.r.flag |= CARRY;
            if (dest.value & 0x80L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            wrapflag = 0;
            regs.r.flag &= CARRYOFF;
            if (dest.value & 0x80L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value = ((dest.value >> 1) & 0x7fL) + wrapflag;
         }
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      while (temp--)
         {
         if (dest.value & 1)
            {
            wrapflag = SIGNEDW;
            regs.r.flag |= CARRY;
            if (dest.value & 0x8000L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            wrapflag = 0;
            regs.r.flag &= CARRYOFF;
            if (dest.value & 0x8000L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value = ((dest.value >> 1) & 0x7fffL) + wrapflag;
         }
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void sahf()
   {
   regs.r.flag &= 0xff00;
   regs.r.flag |= regs.r.ax.byte.ah;
   }

void sar()
   {
   unsigned int temp,signflag;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (temp)
      regs.r.flag &= OVERFLOWOFF;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         if (dest.value & 1)
            regs.r.flag |= CARRY;
         else
            regs.r.flag &= CARRYOFF;
         signflag = dest.value & 0x80L;
         dest.value = ((dest.value >> 1) & 0x7fL) + signflag;
         }
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
     else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      while (temp--)
         {
         if (dest.value & 1)
            regs.r.flag |= CARRY;
         else
            regs.r.flag &= CARRYOFF;
         signflag = dest.value & 0x8000L;
         dest.value = ((dest.value >> 1) & 0x7fffL) + signflag;
         }
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void sbb()
   {
   UNSIGNED_LONG carryvalue;

   if (regs.r.flag & CARRY)
      carryvalue = 0;   /* complemented borrow */
   else
      carryvalue = 1;
   dest.value += ~source.value + carryvalue;
   if (instruction.flags & BYTE)
      {
      ovcarry(((unsigned int)~source.value),(unsigned int)carryvalue,BYTE);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      ovcarry(((unsigned int)~source.value),(unsigned int)carryvalue,WORD);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void scasb()
   {
   unsigned int inc;

   if (regs.r.flag & DIRECTION)
      inc = -1;
   else
      inc = 1;
   if (prefix & REPZERO)
      {
      prefix |= ISZFON;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchbyte(regs.r.es,regs.r.di);
         dest.oldvalue = dest.value;
         dest.value = regs.r.ax.byte.al - dest.value;
         if (memok)
            {
            regs.r.di += inc;
            source.value = ~dest.oldvalue;
            dest.oldvalue = regs.r.ax.byte.al;
            ovcarry((unsigned int)source.value,1,BYTE);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(BYTE);
            }
         }
      else
         prefix = 0;
      }
   else
      if (prefix & REPNOTZERO)
         {
         prefix |= ISZFOFF;
         if (regs.r.cx.word)
            {
            dest.value = (UNSIGNED_LONG)fetchbyte(regs.r.es,regs.r.di);
            dest.oldvalue = dest.value;
            dest.value = regs.r.ax.byte.al - dest.value;
            if (memok)
               {
               regs.r.di += inc;
               source.value = ~dest.oldvalue;
               dest.oldvalue = regs.r.ax.byte.al;
               ovcarry((unsigned int)source.value,1,BYTE);
               regs.r.flag ^= CARRY;
               regs.r.flag ^= AUXCARRY;
               flags(BYTE);
               }
            }
         else
            prefix = 0;
         }
      else
         {
         dest.value = (UNSIGNED_LONG)fetchbyte(regs.r.es,regs.r.di);
         dest.oldvalue = dest.value;
         dest.value = regs.r.ax.byte.al - dest.value;
         if (memok)
            {
            regs.r.di += inc;
            source.value = ~dest.oldvalue;
            dest.oldvalue = regs.r.ax.byte.al;
            ovcarry((unsigned int)source.value,1,BYTE);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(BYTE);
            }
         }
   }

void scasw()
   {
   unsigned int inc;

   if (regs.r.flag & DIRECTION)
      inc = -2;
   else
      inc = 2;
   if (prefix & REPZERO)
      {
      prefix |= ISZFON;
      if (regs.r.cx.word)
         {
         dest.value = (UNSIGNED_LONG)fetchword(regs.r.es,regs.r.di);
         dest.oldvalue = dest.value;
         dest.value = regs.r.ax.word - dest.value;
         if (memok)
            {
            regs.r.di += inc;
            source.value = ~dest.oldvalue;
            dest.oldvalue = regs.r.ax.word;
            ovcarry((unsigned int)source.value,1,WORD);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(WORD);
            }
         }
      else
         prefix = 0;
      }
   else
      if (prefix & REPNOTZERO)
         {
         prefix |= ISZFOFF;
         if (regs.r.cx.word)
            {
            dest.value = (UNSIGNED_LONG)fetchword(regs.r.es,regs.r.di);
            dest.oldvalue = dest.value;
            dest.value = regs.r.ax.word - dest.value;
            if (memok)
               {
               regs.r.di += inc;
               source.value = ~dest.oldvalue;
               dest.oldvalue = regs.r.ax.word;
               ovcarry((unsigned int)source.value,1,WORD);
               regs.r.flag ^= CARRY;
               regs.r.flag ^= AUXCARRY;
               flags(WORD);
               }
            }
         else
            prefix = 0;
         }
      else
         {
         dest.value = (UNSIGNED_LONG)fetchword(regs.r.es,regs.r.di);
         dest.oldvalue = dest.value;
         dest.value = regs.r.ax.word - dest.value;
         if (memok)
            {
            regs.r.di += inc;
            source.value = ~dest.oldvalue;
            dest.oldvalue = regs.r.ax.word;
            ovcarry((unsigned int)source.value,1,WORD);
            regs.r.flag ^= CARRY;
            regs.r.flag ^= AUXCARRY;
            flags(WORD);
            }
         }
   }

void segcs()
   {
   }

void segds()
   {
   }

void seges()
   {
   }

void segss()
   {
   }

void shl()
   {
   unsigned int temp;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         if (dest.value & 0x80L)
            {
            regs.r.flag |= CARRY;
            if (dest.value & 0x40L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            regs.r.flag &= CARRYOFF;
            if (dest.value & 0x40L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value <<= 1;
         }
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
     else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      while (temp--)
         {
         if (dest.value & 0x8000L)
            {
            regs.r.flag |= CARRY;
            if (dest.value & 0x4000L)
               regs.r.flag &= OVERFLOWOFF;
            else
               regs.r.flag |= OVERFLOW;
            }
         else
            {
            regs.r.flag &= CARRYOFF;
            if (dest.value & 0x4000L)
               regs.r.flag |= OVERFLOW;
            else
               regs.r.flag &= OVERFLOWOFF;
            }
         dest.value <<= 1;
         }
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void shr()
   {
   unsigned int temp;

   if (instruction.flags & (NOCOUNT+COUNT))  /* if no flags, then immediate */
      {
      if (instruction.flags & NOCOUNT)
         temp = 1;
      else
         temp = regs.r.cx.byte.cl;
      }
   else
      temp = source.value;
   if (instruction.flags & BYTE)
      {
      while (temp--)
         {
         if (dest.value & 1)
            regs.r.flag |= CARRY;
         else
            regs.r.flag &= CARRYOFF;
         if (dest.value & 0x80L)
            regs.r.flag |= OVERFLOW;
         else
            regs.r.flag &= OVERFLOWOFF;
         dest.value = (dest.value >> 1) & 0x7fL;
         }
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
     else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      while (temp--)
         {
         if (dest.value & 1)
            regs.r.flag |= CARRY;
         else
            regs.r.flag &= CARRYOFF;
         if (dest.value & 0x8000L)
            regs.r.flag |= OVERFLOW;
         else
            regs.r.flag &= OVERFLOWOFF;
         dest.value = (dest.value >> 1) & 0x7fffL;
         }
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void stc()
   {
   regs.r.flag |= CARRY;
   }

void std()
   {
   regs.r.flag |= DIRECTION;
   }

void sti()
   {
   regs.r.flag |= INTERRUPT;
   interruptcount = 1;
   }

void stosb()
   {
   unsigned int inc;

   if (regs.r.flag & DIRECTION)
      inc = -1;
   else
      inc = 1;
   if (prefix & REP)
      {
      prefix |= FORGETZF;
      if (regs.r.cx.word)
         {
         putbyte(regs.r.es,regs.r.di,((unsigned int)regs.r.ax.byte.al));
         if (memok)
            regs.r.di += inc;
         }
      else
         prefix = 0;
      }
   else
      {
      putbyte(regs.r.es,regs.r.di,((unsigned int)regs.r.ax.byte.al));
      if (memok)
         regs.r.di += inc;
      }
   }

void stosw()
   {
   unsigned int inc;

   if (regs.r.flag & DIRECTION)
      inc = -2;
   else
      inc = 2;
   if (prefix & REP)
      {
      prefix |= FORGETZF;
      if (regs.r.cx.word)
         {
         putword(regs.r.es,regs.r.di,((unsigned int)regs.r.ax.word));
         if (memok)
            regs.r.di += inc;
         }
      else
         prefix = 0;
      }
   else
      {
      putword(regs.r.es,regs.r.di,((unsigned int)regs.r.ax.word));
      if (memok)
         regs.r.di += inc;
      }
   }

void sub()
   {
   dest.value -= source.value;
   if (instruction.flags & BYTE)
      {
      if (dest.value > 0xffL)
         regs.r.flag &= CARRYOFF;
      else
         regs.r.flag |= CARRY;
      ovcarry(((unsigned int)~source.value),1,BYTE);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      if (dest.value > 0xffffL)
         regs.r.flag &= CARRYOFF;
      else
         regs.r.flag |= CARRY;
      ovcarry(((unsigned int)~source.value),1,WORD);
      regs.r.flag ^= CARRY;
      regs.r.flag ^= AUXCARRY;
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   }

void test()
   {
   dest.value &= source.value;
   if (instruction.flags & BYTE)
      flags(BYTE);
   else
      flags(WORD);
   regs.r.flag &= CARRYOFF;
   regs.r.flag &= OVERFLOWOFF;
   }

void wait()
   {
   memok = wait_status;  /* set appropriate break status */
/*   regs.r.ip--;  after an interrupt, the processor returns to the wait
                   instruction.  This is no longer used, as WAIT is either
                   a NOP (never wait) or a breakpoint (never execute). */
   }

void xchg()  /* I think this is the only place the
                source.where values are used */
   {
   if (instruction.flags & BYTE)
      {
      if (dest.where & REG)
         regs.bytearray[dest.reg] = source.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)source.value));
      if (source.where & REG)
         regs.bytearray[source.reg] = dest.value;
      else
         putbyte(source.seg,source.ofs,((unsigned int)dest.value));
      }
   else
      {
      if (dest.where & REG)
         regs.wordarray[dest.reg] = source.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)source.value));
      if (source.where & REG)
         regs.wordarray[source.reg] = dest.value;
      else
         putword(source.seg,source.ofs,((unsigned int)dest.value));
      }
   }

void xlat()
   {
   unsigned int sourceseg,tempofs,temp;

   if (prefix & SEG)
      switch (prefix & SEG)
         {
         case CSSEG:
            sourceseg = regs.r.cs;
            break;
         case DSSEG:
            sourceseg = regs.r.ds;
            break;
         case ESSEG:
            sourceseg = regs.r.es;
            break;
         case SSSEG:
            sourceseg = regs.r.ss;
            break;
         }
   else
      sourceseg = regs.r.ds;
   tempofs = regs.r.bx.word + regs.r.ax.byte.al;
   temp = fetchbyte(sourceseg,tempofs);
   if (memok)
      {
      regs.r.ax.byte.al = temp;
      }
   }

void xor()
   {
   dest.value ^= source.value;
   if (instruction.flags & BYTE)
      {
      flags(BYTE);
      if (dest.where & REG)
         regs.bytearray[dest.reg] = dest.value;
      else
         putbyte(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   else
      {
      flags(WORD);
      if (dest.where & REG)
         regs.wordarray[dest.reg] = dest.value;
      else
         putword(dest.seg,dest.ofs,((unsigned int)dest.value));
      }
   regs.r.flag &= CARRYOFF;
   regs.r.flag &= OVERFLOWOFF;
   }

/*
   This is where the array of pointers to functions is set up.  Note that
   the order is the SAME as that of the opnames array, and that the
   indexing is keyed to the defines in OPDEFS.H.
   THEY MUST BE KEPT THAT WAY!  IF YOU CHANGE ANY OF THEM YOU MUST CHANGE
   THEM ALL!

   That the order is the same as the order of the routines above reflects
   bookkeeping while writing the routines.  They do not have to be in
   order, but think of the poor programmer after you!
*/

void (*opfns[113])() = {
db,    aaa,   aad,   aam,
aas,   adc,   add,   and,
bound, call,  callf, cbw,
clc,   cld,   cli,   cmc,
cmp,   cmpsb, cmpsw, cwd,
daa,   das,   dec,   div,
enter, esc,   hlt,   idiv,
imul,  in,    inc,   insb,
insw,  intr,  into,  iret,
ja,    jae,   jcxz,  jg,
jge,   jmp,   jmpf,  jna,
jnae,  jng,   jnge,  jno,
jns,   jnz,   jo,    jpe,
jpo,   js,    jz,    lahf,
lds,   lea,   leave, les,
lock,  lodsb, lodsw, loop,
loopz, loopnz,mov,   movsb,
movsw, mul,   neg,   nop,
not,   or,    out,   outsb,
outsw, pop,   popa,  popf,
push,  pusha, pushf, rcl,
rcr,   repz,  repnz, ret,
retf,  rol,   ror,   sahf,
sar,   sbb,   scasb, scasw,
segcs, segds, seges, segss,
shl,   shr,   stc,   std,
sti,   stosb, stosw, sub,
test,  wait,  xchg,  xlat,
xor
   };
