/*
   This file is copyright (c) 1986, 1987 by William Couture.  All rights reserved.
*/


#include <compiler.h>  /* Makes the source code more transportable */
#include <vim.h>
#include <regstruc.h>
#include <opdefs.h>
#include <machine.h>  /* for a PC-XT or a Z150?  */
#include <include\ctype.h>

#define PC 1  /* processor seems to be an 8088/8086 */
#define AT 0  /* processor seems to be an 80286 */

#define WHERE_SIZE 0x7ff   /* size of "where was I?" queue */


extern void (*opfns[113])();  /* the functions to simulate the instructions */

extern char regnames[23][3];  /* register names (i.e. AX, BX, CH, etc) */
extern char opnames[113][7];  /* names of instructions, used for X command */
extern char flagnames[2][9][3];  /* names of flags, for RF command */

extern char *helptext[HELPSIZE];  /* the help text */

/* char alignment_byte_1;  used to word align data */
    /* NOTE:  This byte is moved as necessary for word alignment of data */
unsigned char byte_one;  /* the 1st byte of the decoded instruction */
unsigned char port_61_save;  /* make sure speaker is off in VIM */
unsigned int aadaam_val;  /* value to use for AAD/AAM (not necessarily 0ah) */

union regunion regs;  /* See the regstruc.h file for details */

struct optable instruction;  /* the instruction as decoded for program
                                internal use */

struct tofrom dest,source;   /* info about the source and dest fields in the
                                current instruction, supplied by decode()
                                and decodens() */

struct addrtrap brkpoints[21];  /* list of watched memory addresses.  See
                                   vim.h for info.  NOTE:  the used entries
                                   are 1..20, as numbreaks = 0 means that
                                   no breakpoint have been set! */
struct addrtrap sticky[10];  /* store up to 10 sticky breakpoints */
unsigned int brkcounts[21],stickycounts[10];  /* allow countdown on breakpoints */

struct intrptr int8ptr,int9ptr,intaptr,intbptr,intcptr,intdptr,inteptr,intfptr,
      int70ptr,int71ptr,int72ptr,int73ptr,int74ptr,int75ptr,int76ptr,int77ptr,
      ctrlcptr;
 /* interrupt pointers, for internal use to watch the hardware interrupts,
    and take over the control-c interrupt handler */

/* Since, on the PC, the NMI is used for Parity error detection, and I
   don't want VIM to keep running if memory fails, I will not be saving/
   watching the NMI vector. */

int mod,reg,rm;  /* EA byte fields, set by decode() and decodens() */

unsigned int numbreaks;  /* number of brkpoints set */
unsigned int prefix;  /* global info on the current prefix instructions */
unsigned int memok = TRUE;  /* if TRUE, have not referenced a guarded address */
unsigned int errmessage;  /* set on HLT, to let hardware interrupts bypass
                             the halt */
unsigned int wait_status = TRUE;  /* TRUE if WAIT is a NOP, else false */
unsigned int interruptcount; /* interrupts off while interruptcount, due to
                                mov/pop into segreg */
unsigned int segoffset;  /* virtual relocation segment */
unsigned int rom_offset;  /* used for ROM BIOS replacement */
int rom_count;  /* count of replaced segments */
unsigned int rom_dma = FALSE;  /* a flag to remap DMA to and from remapped ROM */
unsigned int rom_segs[4];  /* where the should be */
unsigned int rom_maps[4];  /* and where they are */
unsigned int int_passing = FALSE;  /* pass or ignore unused interrupts? */
unsigned int prefixip;  /* addr of last prefix instruction of current
                         instruction, used in case of an "interrupt" during
                         multi-prefixed instructions to emulate the 8088's
                         pushing the addr of the last prefix. See decode.c */
unsigned int start_ip;  /* addr of start of instruction, including prefixes.
                           This is used for interrupts caused by the BOUND
                           instruction */
unsigned int pcscreen = FALSE;  /* flag to allow screen swapping */
unsigned int pclone = FALSE;  /* flag to turn VIM into a PC-XT clone.  This allows
                         I/O to the PC's ports, with translation of
                         the DMA addresses (ports 0x81 - 0x84) and handling
                         of keyboard port 0x60 (i.e. save before BIOS
                         resets it). */
unsigned char port60queue[64];  /* port 60 values for keyboard interrupt
                                   queue.  NOTE THAT THESE ARE CHARs! */
unsigned int port60ptr;  /* pointer to next free element of port60queue */
unsigned int hotkeypressed;  /* Toggle from simulated 8088 to VIM by a
                                hot key, this is set TRUE when it is
                                pressed (rightshift-ESC).  THIS IS NOT
                                PASSED TO THE SIMULATED PROGRAM! */
unsigned int rshiftstatus;  /* flag containing the status of the right shift
                               key.  TRUE=pressed, FALSE=not pressed */
unsigned int rshiftused;  /* Has right shift been passed to interrupt queue?
                             Used to keep hotkey shift interrupt from
                             simulator. */
unsigned int lshiftstatus;  /* flag containing the status of the left shift
                               key.  TRUE=pressed, FALSE=not pressed */
unsigned int lshiftused;  /* Has left shift been passed to interrupt queue?
                             Used to keep hotkey shift interrupt from
                             simulator. */
unsigned int keybreak;  /* remember break status of new IBM keyboard scan
                           codes */
unsigned int keybdoff = FALSE;  /* flag to turn off keyboard */
unsigned int hkscancode = 1;  /* user definable scancode for the hotkey */
unsigned int hwintflag = FALSE;  /* flag to simulate a 8259 interrupt */
unsigned int hw8259on = FALSE;  /* 8259 needs to be reset by OUTing 0x20 or
                            0x6? to port 0x20.  Until then, 8259 is OFF. */
unsigned int hwintnumber;  /* int. number "from" the "8259" */
unsigned int hwintstack[16];  /* pending interrupts from the "8259" */
unsigned int irq_divisors[16];  /* irq divisor values */
unsigned int irq_counters[16];  /* irq counter values */
unsigned int clockcount;  /* use to slow down simulator clock interrupt */
unsigned int clockdivide;  /* use to adjust for PC or AT speed */
unsigned int pspaddr;  /* used to hold the Program Segment Prefix of this
                          program, to make a local copy of DOS */
unsigned int codesegment;  /* the programs code segment */
unsigned int datasegment;  /* the programs data segment */
unsigned int highsegment;  /* the value of the higher of the two segments */
unsigned int maxpara;  /* maximum available paragraphs in virtual memory */
unsigned int memmap[16];  /* values for virtual segment mapping */
unsigned int dmamap[16];  /* values for DMA segment mapping */
unsigned int eoiwait = FALSE;  /* true if waiting for an EOI.  Used for checking
                          for the end of the AT keyboard routine */
unsigned int breakonio = FALSE;  /* flag to cause break on I/O */
unsigned int portbreaklist[21];  /* list of ports to break on */
unsigned int portbreaknum;  /* number of ports in list */

unsigned int patchseg[21];
unsigned int patchofs[21];
unsigned int patchbyte[21];
unsigned int patched = FALSE;
             /* location, value, and flag for patching Pre-DOS 3.0 */

unsigned int stickybreaks = FALSE;  /* sticky breakpoints enabled/disabled */
unsigned int numsticky = 0;  /* number of sticky breakpoints set */
unsigned int deadmem;  /* segment for mapping of non-existant memory */
unsigned int fake_retrace = FALSE;  /* flag for faking RETRACE */
unsigned int retr_count;  /* a count for the INT 8h handler */
unsigned int retr_num;  /* number of ticks per on/off for retrace */
unsigned int retr_flag = FALSE;  /* flag saying if retrace is on or off */
unsigned int port64hold;  /* a keyboard flag */

UNSIGNED_LONG where_count;  /* count of instructions executed */
unsigned int whereptr;  /* pointer into "where was I?" queue */
unsigned int wherequeueofs[2048];  /* queue of offsets for where? */
unsigned int wherequeueseg[2048];  /* queue of segments for where? */
unsigned int cmdptr;  /* pointer into the command line from the user */
unsigned int fileinput = 0;  /* file handle for alternate input */
unsigned int outfile;  /* file handle for alternate output */
unsigned int outputaborted;  /* the output echoing was aborted */
unsigned int doserrnum;  /* error number for DOS */
unsigned int zf_file;  /* file number for ZF command */

UNSIGNED_LONG zf_pos, lseek();  /* file position and lseek define */

unsigned char cmdline[150];  /* command line from user interface */
unsigned char globalstr[150];  /* global temporary string */
unsigned char tempstr[150];  /* another temporary string */
unsigned char stickylist[220];  /* list of sticky breakpoints */
unsigned char outfilename[150];  /* the name for alternate output */
unsigned char infilename[150];  /* the name for alternate output */
unsigned char screen[12290];  /* screen save area */
unsigned char zf_screen[1850];  /* 1840 plus slop */
unsigned char zf_name[150];
unsigned char vga_norm_dac[768];  /* VGA normal mode */
unsigned char vga_virt_dac[768];  /* VGA virtual mode */


unsigned int videobase;  /* base address for video registers */
unsigned int feature_addr;  /* videobase+6, used for screen swapping */
int egascreen;  /* pointer into video mode table */
int vgascreen;  /* say that VGA is installed */
unsigned int hercules;  /* is a Hercules video adapter present? */
unsigned int attribflipflop = 1;
       /* current state (0=address, 1=data) of port 3C0 */
unsigned int mdap3b8;  /* default state of port 3B8 (MDA) */
unsigned int mdap3bf;  /* default state of port 3BF (Hercules) */
unsigned int cgap3d8;  /* default state of port 3D8 (CGA) */
unsigned int cgap3d9;  /* default state of port 3D9 (CGA) */
/* unsigned int cursor_start;  starting line for changing cursor
unsigned int cursor_flag;  */
unsigned char mdaports[8];  /* contents of the MDA ports 3B0-3BF */
unsigned char cgaports[6];  /* contents of the CGA ports 3D0-3DB */
unsigned char egaports[28];  /* the ega/vga ports, see VIM.H for layout */
unsigned char mdanorm[NUM_CRT_REG];
unsigned char mdaregs[NUM_CRT_REG];
      /* registers to be changed within VIM */
unsigned char mdamode[18] = {0x61,0x50,0x52,0x0f,0x19,0x06,0x19,0x19,0x02,
                             0x0d,0x0b,0x0c,0,0,0,0,0,0};
     /* registers to restore "normal" mode */
unsigned char cganorm[NUM_CRT_REG];
unsigned char cgaregs[NUM_CRT_REG];
     /* registers to be changed within VIM */
unsigned char cgamode[18] = {0x71,0x50,0x5a,0x0a,0x1f,0x06,0x19,0x1c,0x02,
                             0x07,0x06,0x07,0,0,0,0,0,0};
     /* registers to restore "normal" mode */

unsigned char sinorm[NUM_SI_REG];
unsigned char siregs[NUM_SI_REG];  /* si registers watched within VIM */

unsigned char attribnorm[NUM_ATTRIB_REG];
unsigned char attribregs[NUM_ATTRIB_REG];  /* attribute regs watched within VIM */

unsigned char seqnorm[NUM_SEQ_REG];
unsigned char seqregs[NUM_SEQ_REG];  /* sequencer regs watched within VIM */

unsigned int seq_reg_count = NORM_SEQ_REG;
unsigned int si_reg_count = NORM_SI_REG;
unsigned int attrib_reg_count = NORM_ATTRIB_REG;
unsigned int crt_reg_count = NORM_CRT_REG;
unsigned int extended_swap = FALSE;

/* unsigned char egamode[2][64]; */ /* data to be read from EGA ROM's */

unsigned int mixbytereg[8] =
   {
/* al  cl  dl  bl  ah  ch  dh  bh   r/m byte reg definitions 0..7 */
    0,  2,  4,  6,  1,  3,  5,  7
/* al  ah  cl  ch  dl  dh  bl  bh   regs.bytearray register order  */
   };

unsigned int flagmasks[9] =
   {
   OVERFLOW,DIRECTION,INTERRUPT,TRAP,SIGN,ZERO,AUXCARRY,PARITY,CARRY
   };  /* used in MOREVIM.C to display and modify the flags */

char tab[2] = "\t";


#ifdef ZENITH
unsigned char scratchmem[16386];  /* a 16K bank in the Zenith at F000:0000 */
#endif

unsigned char dumpblank[80] =  /* blank for the DUMP command */
"    :                             -                                          ";
unsigned char err_num[10] = "number?\n";  /* error messages for */
unsigned char err_str[10] = "string?\n";  /* user interface */
unsigned char err_cmd[12] = "command?\n";
unsigned char err_fn[12] =  "filename?\n";
unsigned char err_rng[10] = "range?\n";
unsigned char nl_str[2] = "\n";  /* used frequently */
unsigned char space_str[2] = " ";  /* used mostly in showvideoregs() */
unsigned char filename[150];  /* allow for paths */
unsigned char imagename[150];
char tabstr[] = "                       \t      \t";
                         /* blank for displaying a disassembled instruction */

/* The information needed by the checkaddress function pointer */

void (*checkaddress)();  /* the pointer itself */
void checkaddress1(),checkaddress2();  /* the functions it points to,
             without and with re-direction to saved screen memory */
void (*checkwaddress)();  /* the pointer itself */
void checkwaddress1(),checkwaddress2();  /* the functions it points to,
             without and with re-direction to saved screen memory */
UNSIGNED_LONG min_screen,max_screen;

/* "Local" data for the menu() routine */

unsigned int asmspacing,asmip,asmcs,bytectr;  /* for ASM */
unsigned int byte,flag,count,keyin;
unsigned int unasmip,unasmcs,unasmlen,tempip,holdip,holdcs;
unsigned int dumpseg,dumpofs,dumplen;
unsigned int dotrace,countflag,instname,xchar;  /* for the X command */
UNSIGNED_LONG templong,longnum,temp,temp2,temp3; /* temp2, temp3 unused? */
int regindex,valid;
unsigned char hexstr[20],name[8],xname[220];
char *hexword();

char zf_header[] = "Find, Again, Esc to quit, cursor keys active: PgUp, PgDn, Home, End, Up, Down\n-----\n";
char copyright[] = "Copyright (c) 1988 by William Couture.  All rights reserved. ";

/* unsigned char byte_one;  the 1st byte of the decoded instruction */
    /* NOTE:  This byte is moved as necessary for word alignment of data */

void setmaps(),setup(),cleanup(),menu(),get_patches(),skip_cmd_spaces();
void onoroff(),onoroff(),showwhere(),regdisplay(),putsinstr(),showvideoregs();
void godoit(),execute(),readint();


main(argc,argv)
   int argc;
   char *argv[];
   {
   unsigned int segregs[4];

#asm
  ;
  ;  This NOP exists for alignment of the code segment.  It seems to
  ;  make a noticable (~2%) difference in the interpreter speed.
  ;  If making a change slows the interpreter down, try including/removing
  ;  this byte of code.
  ;
  ;  Other NOPs have also been placed in HELP.C and DECODENS.ASM
  ;
   nop    ;Code alignment NOP
#endasm
   /* make sure this is DOS 2.0 or greater */
   cmdptr = 1;  /* use cmdptr as a temporary variable, start at 1 (pre-2.0) */
#asm
   mov ah,30h    ;function = 30h, DOS version
   int 21h       ;call DOS
   cmp al,2      ;2.0 or above?
   jnae wrongdos ;no, skip this
   mov word ptr cmdptr_,0  ;2.0 or greater, everything ok
wrongdos:
#endasm
   if (cmdptr)  /* if pre-2.0 */
      {
      PUTS("Sorry, but VMDEBUG requires DOS 2.0 or above.\n");
      exit();
      }
   segread(segregs);
   datasegment = segregs[2];  /* get data and code seg.  Note implied small
                                 memory model. */
   codesegment = segregs[0];

#ifdef DESMET
   POKE( (datasegment & 0xff),codesegment,9);
   POKE( ((datasegment >> 8) & 0xff),codesegment,0x0a);
          /* Put the location of the data segment in a place that is
             available if ONLY the code segment is known.  These bytes are
             the first two NOP's above, and are used for the interrupt
             handler routines.  See below for a full explanation!

             WARNING:  DESMET SPECIFIC!!!!!
           */
   pspaddr = wpeek(codesegment,1) + 3;
   pspaddr = wpeek(codesegment,pspaddr+3);
   pspaddr = wpeek(datasegment,pspaddr);
#else
   pspaddr = codesegment - 0x10;  /* hope this works! */
#endif

/*
   The comments for the above were moved to the file EXPLAIN.H, to shorten
   the working length of the file VIM.C.
*/
   highsegment = datasegment;
   if (codesegment > datasegment)  /* do not assume DS > CS */
      highsegment = codesegment;
   segoffset = highsegment + 0x1000;   /* allow for 64K (small memory model) */
   if (segoffset & 0x0fff)
      segoffset = (segoffset & 0xf000) + 0x1000;
      /* NOTE:  It is IMPORTANT that the segment offset relocation of the
         virtual program space be a multiple of 64K (i.e. have only the
         upper 4 bits of the segment set) so that the DMA (for disk access)
         can be relocated into the virtual address space by intercepting
         and relocating the calls to I/O ports 0x81-8x83 (corresponding
         to DMA channels 2,3,1 respectively), which set the upper 4 bits
         for a 20-bit DMA address.
       */
   regs.r.cs =
   regs.r.ds =
   regs.r.es =
   regs.r.ss = 0;        /* make sure segs start at 0 */
   regs.r.ip = 0x100;  /* let programs think they have a psp */
   port_61_save = INBPORT(0x61);  /* get default info */
   videobase = wpeek(0x40,0x63);
   feature_addr = videobase + 6;
   if (videobase < 0x3d0)
      {
      min_screen = 0xb0000L;
      max_screen = 0xb0fffL;
      }
   else
      {
      min_screen = 0xb8000L;
      max_screen = 0xb8fffL;
      }
   checkaddress = checkaddress1;  /* default is no video screen */
   checkwaddress = checkwaddress1;
   deadmem = 0xd000;  /* default to mapping "dead" memory to segment D000 */
   if (argc == 1)
      {
#asm
   mov ax,3     ;force CGA screen
   cmp word ptr videobase_,3d0h   ;is this mono?
   jae itiscolor              ;no, leave it CGA
   mov ax,7     ;force MDA scren
itiscolor:
   push bp      ;save BP, just in case BIOS doesn't
   int 10h      ;call BIOS
   pop bp       ;restore BP
#endasm
      }
   if ((wpeek(0,0x0108)) || (wpeek(0,0x10a)))  /* if INT 42 is used */
      if ((wpeek(0,0x04a8)) || (wpeek(0,0x4aa)))  /* if EGA save pointer is used */
         if ((wpeek(0xc000,0x1e) == 0x4249) && (PEEK(0xc000,0x20) == 'M'))
                               /* standard (UGH!) test for EGA adapter */
            if (((cmdptr = PEEK(0,0x488)) & 0x01) || ((cmdptr & 0x07) > 4))
                        /* if EGA is primary */
               {
               if (videobase == 0x3d4)  /* color EGA is primary display */
                  {
                  if ((PEEK(0,0x488) & 0x0f) == 9)  /* if enhanced */
                     egascreen = 0x16;  /* table number to use, not mode */
                  else
                     egascreen = 3;
                  }
               else  /* monochrome */
                  egascreen = 7;
               }
   cmdptr = 0;   /* assume EGA */
#asm
   mov ax,1c00h            ;get save size
   mov cx,5                ;save regs and DAC
   push bp                 ;always save bp
   int 10h                 ;call BIOS
   pop bp                  ;restore bp
   cmp al,1ch              ;VGA present?
   jne vga_not_installed   ;no, skip this
   mov word ptr cmdptr_,bx ;size of save needed
vga_not_installed:
#endasm
   if (cmdptr)   /* VGA installed */
      {
      vgascreen = 1;
      attrib_reg_count++;  /* one more for VGA */
      egascreen = -egascreen;  /* turn off EGA for now */
      }
   mdaports[P3B8] = 0x29;  /* save video information */
   mdaports[P3B9] = 0;
   mdaports[P3BF] = 0;
   cgaports[P3D8] = 0x29;
   cgaports[P3D9] = 0;
   mdap3b8 = mdaports[P3B8];
   mdap3bf = mdaports[P3BF];
   cgap3d8 = cgaports[P3D8];
   cgap3d9 = cgaports[P3D9];
   setvideoregs();  /* initalize stored video mode */
   if (ispcorat() == PC)
      clockdivide = 2000;  /* divide clock interrupt by 2000 to give about 1
                              tick every 110 seconds on a PC.  This minimizes
                              the overhead spent in clock processing.
                              This may have to be adjusted for best results.
                            */
   else
      clockdivide = 500;  /* figure an AT is about 4 times as fast */
   clockcount = 0;  /* count of clock ticks for divisor */
   filename[0] = '\0';  /* the filename to be loaded */
/*
   Save the hardware interrupt vectors.  They will be replaced at runtime
   by calls to C routines that will trap the needed information for the
   simulator when doing PCLONE simulation.
*/
   readint(&int8ptr,0x08);  /* Int 8, timer */
   readint(&int9ptr,0x09);  /* Int 9, keyboard */
   readint(&intaptr,0x0a);  /* Int 0x0a, reserved */
   readint(&intbptr,0x0b);  /* Int 0x0b, Serial Port #2 */
   readint(&intcptr,0x0c);  /* Int 0x0c, Serial Port #1 */
   readint(&intdptr,0x0d);  /* Int 0x0d, Hard disk attention */
   readint(&inteptr,0x0e);  /* Int 0x0e, Floppy Disk attention */
   readint(&intfptr,0x0f);  /* Int 0x0f, Parallel port */
   readint(&int70ptr,0x70);  /* IRQ8, real time clock */
   readint(&int71ptr,0x71);  /* IRQ9, redirected to Int 0x0a */
   readint(&int72ptr,0x72);  /* IRQ10, reserved */
   readint(&int73ptr,0x73);  /* IRQ11, reserved */
   readint(&int74ptr,0x74);  /* IRQ12, reserved */
   readint(&int75ptr,0x75);  /* IRQ13, coprocessor */
   readint(&int76ptr,0x76);  /* Int 0x76, Hard disk attention (IBM-AT) */
   readint(&int77ptr,0x77);  /* IRQ15, reserved */
   readint(&ctrlcptr,0x23);  /* save the original ctrl-c handler */

   setmaps();  /* set map data for virtual memory and DMA mapping */
   setup();  /* set up for processing */
   menu();  /* go do the simulator */
   cleanup();  /* reset everything before exiting */
   }

void setmaps()
   {
   unsigned int memtop,memused,count,dmaseg;

   memtop = wpeek(0,0x413);  /* get system memory */
   memtop = ((memtop >> 6) & 0x000f) + ((memtop & 0x3f) ? 1:0);
                                                  /* get in low nibble */
   memused = ((segoffset+rom_offset) >> 12) & 0x000f;
                                      /* get number of 64K blocks used */
   maxpara = (memtop - memused) * 0x1000;  /* number of paragraphs in
                                              virtual memory */
   dmaseg = deadmem >> 12;  /* find DMA high 4 bits */
   for (count = 0; count < 16; count++)
      {
      if (count >= (memtop - memused))
         {
         if (count > 9)
            {
            memmap[count] = 0;  /* direct memory mapping */
            dmamap[count] = 0;  /* direct DMA mapping */
            }
         else
            {
            memmap[count] = deadmem - (0x1000 * count);
                                           /* map to deadmem:xxxx */
            dmamap[count] = dmaseg - count;
            }
         }
      else
         {
         memmap[count] = segoffset + rom_offset;  /* map to virtual memory */
         dmamap[count] = memused;
         }
      }
   if (rom_offset)
      for (count = 0; count < rom_count; count++)
         {
         memmap[(rom_maps[count] >> 12)] = segoffset + (count << 12) - rom_maps[count];
         if (rom_dma)
            dmamap[(rom_maps[count] >> 12)] = (segoffset >> 12) + count - (rom_maps[count] >> 12);
         }
   }

void setup()  /* right now, this simply sets the Ctrl-C handler */
   {
   void (*func)();  /* declare a pointer to a function.  I prefer to reference
                       things through this pointer than directly through
                       the function name. */
   void ctrlchdlr();  /* declare function so we can set a pointer to it. */

   /* disable interrupts while vectors are being changed */
#asm
   cli    ;NO INTERUPTS!  INLINE ASSEMBLER IS COMPILER SPECIFIC!
#endasm
   func = ctrlchdlr;  /* set up for Int 0x23 (Ctrl-C) handler */
   setmchnvectors(func,0x8c);
       /* and no call locations are needed */

/* IMPORTANT: Turn interrupts back on */
#asm
   sti   ;Interrupts OK now.  INLINE ASSEMBLER IS COMPILER SPECIFIC!
#endasm
   if ((videobase < 0x3d0) && (egascreen == 0))  /* if a regular MDA */
      {                   /* see if this is a Hercules or clone */
#asm
   jmp test_for_herc      ;skip the subroutines
test_herc_mem:
   mov bl,0               ;reset memory present flag
   mov ax,0b800h          ;ax = hercules page 2 segment
   mov es,ax              ;es = hercules page 2 segment
   xor si,si              ;use si for addressing
   mov ax,0abcdh          ;set ax with memory test pattern
   xchg ax,word ptr es:[si]   ;get original data and set test pattern
   xchg ax,word ptr es:[si]   ;replace original data and get test pattern
   cmp ax,0abcdh          ;is the original pattern back?
   jne no_herc_mem        ;no, skip this
   mov bl,1               ;else say memory present
no_herc_mem:
   ret                    ;and quit
test_for_herc:
   xor bx,bx              ;set flags to 0
   mov dx,3bfh            ;Hercules configuration port
   mov al,3               ;enable page 2 and graphics
   out dx,al              ;send it
   call test_herc_mem     ;see if memory is there
   test bl,1              ;is memory present?
   jz no_herc_present     ;no, cannot be hercules card
   xor ax,ax              ;al=0, disable page 2 and graphics
   out dx,al              ;send it
   call test_herc_mem     ;see if memory is there
   xor bl,1               ;if memory is gone, this is a hercules
no_herc_present:
   mov word ptr hercules_,bx  ;set hercules flag
#endasm
      }
   }

void cleanup()
   {
   /* disable interrupts while vectors are being changed */
#asm
   cli    ;NO INTERUPTS!  INLINE ASSEMBLER IS COMPILER SPECIFIC!
#endasm
   resetmchnvectors(&ctrlcptr,0x8c);
/* IMPORTANT: Turn interrupts back on */
#asm
   sti   ;Interrupts OK now.  INLINE ASSEMBLER IS COMPILER SPECIFIC!
#endasm
   redirectinput("con",NO);  /* force input to console */
   redirectoutput("con",NO);  /* force output to console */
   }

void menu()  /* the main workhorse routine of the VIM shell */
   {
   unsigned int holdseg,holdofs,holdofs2,num1,num2,num3;
   char *hexword();

   dumpseg =  /* defaults for dump and unasm */
   unasmcs =
   asmcs = regs.r.cs;
   dumpofs =
   unasmip =
   asmip = regs.r.ip;
#ifdef ZENITH
   PUTS("VMDEBUGZ150 version 1.32B-1   Copyright (c) 1988 by WIlliam Couture.\n");
#else
   PUTS("VMDEBUG version 1.32B-1   Copyright (c) 1988 by William Couture.\n");
#endif
   PUTS("Use \'H\' for Help.\n");
   strcpy(infilename,"vmdebug.cfg");
   redirectinput(infilename,NO);
   do
      {
#asm
   mov ah,0dh     ;flush disk buffers
   int 21h        ;call DOS
#endasm
      if (outputaborted)
         {
         PUTS("Output echoing aborted.\n");
         doserror('W');  /* occured during writing */
         outputaborted = FALSE;
         }
      getstr(">",cmdline);
      cmdptr =
      byte = 0;
      valid =
      flag = TRUE;
      skip_cmd_spaces();
      if (cmdline[cmdptr])
         byte = toupper(cmdline[cmdptr++]);
      switch (byte)
         {
         case 'Z':  /* VIM specific commands */
            parse_z_cmds();
            break;
         case 'I':  /* get a byte from an input port */
            if (getnumber(&longnum))
               {
               if (pclone)
                  {
                  temp = INBPORT((unsigned int)longnum);
                  if ((unsigned int)longnum == videobase+6)
                     attribflipflop = 0;
                  }
               else
                  temp = 0xff;
               PUTS(hexbyte(&temp,globalstr));
               PUTS(nl_str);
               }
            else
               oops(err_num);
            break;
         case 'O':  /* output a byte to a port */
            if ((getnumber(&longnum)) && (getnumber(&temp))
                     && (temp < 0x100L))
               {
               if (pclone)
                  outmap(((unsigned int)longnum),((unsigned int)temp),BYTE);
               if (temp == 0x61L)
                  port_61_save = (unsigned char)longnum;
            /* else toss it in the bit bucket */
               memok = TRUE;
               }
            else
               oops(err_num);
            break;
         case 'V':  /* display virtual screen, if any */
            if (pcscreen)
               {
               setvimscreen();
               while (gchar() > 255)  /* wait for key */
                  ;
               setrealscreen();
               }
            else
               PUTS("Virtual screen not available.\n");
            break;
         case 'N':  /* name of file.  not copied into "psp" */
            skip_cmd_spaces();
            if (cmdline[cmdptr])
               strcpy(filename,strupr(&cmdline[cmdptr]));
            else
               oops(err_fn);
            break;
         case 'L':  /* load file at cs:ip (not .EXE) or after PSP (.EXE) */
            strcpy(xname,cmdline);
            num1 = cmdptr;
            if (parse(globalstr))  /* if Load sectors */
               {
               strcpy(cmdline,xname);
               cmdptr = num1;
               holdseg = regs.r.ds;
               if ((getaddrange(&holdseg,&holdofs,&holdofs2,2)) &&
                            (getnumber(&templong)) && (getnumber(&longnum)))
                  read_sects(holdseg,holdofs,holdofs2,(unsigned int)templong,(unsigned int)longnum);
               else
                  oops(err_num);
               }
            else
               {
               loadfile();
               port60ptr = 0;
               for (count = 0; count < 16; count++)
                  hwintstack[count] = FALSE;  /* no leftover interrupts */
               hwintflag = FALSE;
               where_count = 0;
               unasmcs = regs.r.cs;
               dumpseg = regs.r.ds;
               }
            break;
         case 'C':  /* create file, located at CS:IP, length BX:CX */
            skip_cmd_spaces();
            if (cmdline[cmdptr])
               {
               num1 = cmdptr;
               parse(globalstr);  /* get the argument we know is there */
               valid = parse(globalstr);  /* any more? */
               cmdptr = num1;     /* unparse the command line */
               if (valid)  /* if more than 1 argument */
                  {
                  holdseg = regs.r.ds;   /* check for C [seg:]ofs dr sect num */
                  if ((getaddrange(&holdseg,&holdofs,&holdofs2,2)) &&
                            (getnumber(&templong)) && (getnumber(&longnum)))
                     write_sects(holdseg,holdofs,holdofs2,(unsigned int)templong,(unsigned int)longnum);
                  else
                     oops(err_num);
                  }
               else
                  {
                  strcpy(globalstr,strupr(&cmdline[cmdptr]));
                  createfile();
                  }
               }
            else
               oops(err_fn);
            break;
         case 'H':  /* HELP! */
         case '?':
            if (parse(globalstr))
               {
               num2 = strlen(globalstr);  /* only compute this once */
               strupr(globalstr);  /* force upper case */
               num1 = 0;  /* nothing printed yet */
               for (count = 0; count < HELPSIZE; count++)
                  {
                  if (strncmp(&helptext[count][1],globalstr,num2) == 0)
                     {
                     do
                        {
                        PUTS(&helptext[count++][1]);
                        num1++;
                        if ((helptext[count][0] == '*')
                              && (helptext[count][1] > 'D')
                              && (helptext[count][1] != 'L'))
                           if (strncmp(&helptext[count][1],globalstr,num2) == 0)
                              if (nextpage() == ESC_KEY)
                                 count = HELPSIZE;
                        } while ((count < HELPSIZE) && (helptext[count][0] == '-'));
                     count--;  /* needed to offset the count++ in the loop,
                                  to find the next command */
                     }
                  }
               if (num1 == 0)
                  PUTS("No help available on that subject.\n");
               }
            else
               for (count = 0; count < HELPSIZE; count++)
                  {
                  if (helptext[count][0] == '*')
                     {
                     if (nextpage() == ESC_KEY)
                        count = HELPSIZE;
                     else
                        PUTS(&helptext[count][1]);
                     }
                  else
                     PUTS(&helptext[count][1]);
                  }
            break;
         case 'B':
            skip_cmd_spaces();
            if (cmdline[cmdptr])
               {
               if (toupper(cmdline[cmdptr]) == 'O')
                  {
                  numsticky = 0;
                  stickybreaks = FALSE;
                  PUTS("Sticky breakpoints cleared.\n");
                  }
               else
                  {
                  if (toupper(cmdline[cmdptr]) == 'L')
                     {
                     if (numsticky)
                        {
                        PUTS("Sticky breakpoints: ");
                        PUTS(stickylist);
                        PUTS(nl_str);
                        }
                     else
                        PUTS("No sticky breakpoints.\n");
                     }
                  else
                     {
                     num1 = 10;  /* count down to get 10 breakpoints */
                     numbreaks = 0;  /* start off blank */
                     xname[0] = '\0';
                     set_breakpoints(num1,TRUE);  /* get breakpoints, save list */
                     if ((valid) && (numbreaks))  /* if valid and some entered */
                        {
                        numsticky = numbreaks;
                        for (count = 0; count < numsticky; count++)
                           {
                           sticky[count].lowaddr = brkpoints[count+1].lowaddr;
                           sticky[count].highaddr = brkpoints[count+1].highaddr;
                           stickycounts[count] = brkcounts[count+1];
                           }
                        stickybreaks = TRUE;
                        numbreaks = 0;  /* in case Trace follows this command */
                        PUTS("Sticky breakpoints set.\n");
                        strcpy(stickylist,xname);  /* save list */
                        }
                     }
                  }
               }
            else
               {
               stickybreaks = !stickybreaks;  /* toggle status */
               PUTS("Sticky breakpoints ");
               onoroff(stickybreaks);
               }
            break;
         case 'G':
         case 'P':  /* The Go command */
         case 'J':  /* The J (Go with prefetch watch) command */
            check_for_csip();
            if (valid)
               {
               num1 = (byte != 'G' ? 19:20);  /* how many breakpoints to allow
                                                 the user to set */
               numbreaks = 0;  /* start off blank */
               if ((stickybreaks) && (numsticky))
                  {
                  for (count = 0; count < numsticky; count++)
                     {
                     brkpoints[count+1].lowaddr = sticky[count].lowaddr;
                     brkpoints[count+1].highaddr = sticky[count].highaddr;
                     brkcounts[count+1] = stickycounts[count];
                     }
                  numbreaks += numsticky;
/*                  num1 -= numsticky; */
                  }
               set_breakpoints(num1,FALSE);  /* get breakpoints, no list */
               }
            if (valid)  /* if no bad numbers in breakpoints */
               {
               keyin = 0;  /* just something besides an ESC */
               if (byte == 'P')
                  {
                  count = numbreaks;
                  numbreaks = 0;  /* do not count breaks while finding
                                     the next instruction */
                  holdofs = regs.r.ip;
                  do
                     {
                     decodens();  /* update IP.  Strings are not needed */
                     } while ((instruction.flags & PREFIX)
                                 && ((keyin = gchar()) != ESC_KEY));
                  numbreaks = count + 1;
                  templong = regs.r.cs;
                  brkpoints[numbreaks].lowaddr =
                                    ((templong << 4) + regs.r.ip) & 0xfffffL;
                  brkpoints[numbreaks].highaddr = brkpoints[numbreaks].lowaddr;
                  brkcounts[numbreaks] = 1;  /* only one count for a P */
                  regs.r.ip = holdofs;
                  }
               if (keyin != ESC_KEY)
                  godoit(byte == 'J' ? TRUE : FALSE);
               else
                  numbreaks = 0;
               }
            break;
         case 'X':  /* The eXecute command */
            dotrace = FALSE;  /* default to no Trace style dump */
            longnum = 0x7fffffffL;  /* default to maximum count */
            instname = DBYTE;  /* default to stop on DB instruction */
            xname[0] = '\0';  /* match name defaults to null */
            num3 = 0;  /* length of match name */
            check_for_csip();
            if (valid)
               {
               do
                  {
                  skip_cmd_spaces();
                  count = cmdptr;  /* pointer into command line */
                  xchar = toupper(cmdline[count]);
                  if (xchar == 'T')
                     {
                     cmdptr = count + 1;
                     dotrace = TRUE;  /* yes, do Trace style dump */
                     }
                  if (xchar == 'Q')
                     {
                     cmdptr = count + 1;
                     if ((valid = getnumber(&longnum)) == FALSE)
                        oops(err_num);  /* get count */
                     }
                  if (xchar == 'N')  /* instruction name */
                     {
                     cmdptr = count + 1;
                     parse(globalstr);
                     num1 = 0;
                     strupr(globalstr);  /* force upper case */
                     if (isfillenterstr(globalstr))
                        {
                        num3 = make_inst_string(globalstr,xname);
                        }
                     else
                        {
                        for (num1 = DBYTE + 1; num1 <= XOR; num1++)
                           if (!strcmp(globalstr,opnames[num1]))
                              instname = num1;
                        if (instname == DBYTE)  /* invalid name! */
                           {
                           valid = FALSE;
                           oops(err_str);
                           }
                        }
                     }
                  } while (((xchar == 'T') || (xchar == 'Q') || (xchar == 'N')) && (valid));
               }
            numbreaks = 0;  /* start off blank */
            num1 = 20;
            if ((stickybreaks) && (numsticky))
               {
               for (count = 0; count < numsticky; count++)
                  {
                  brkpoints[count+1].lowaddr = sticky[count].lowaddr;
                  brkpoints[count+1].highaddr = sticky[count].highaddr;
                  brkcounts[count+1] = stickycounts[count];
                  }
               numbreaks += numsticky;
/*               num1 -= numsticky; */
               }
            if (valid)
               set_breakpoints(num1,FALSE);  /* get breakpoints, no list */
            if (valid)  /* if no bad numbers in breakpoints */
               execute(longnum,dotrace,instname,num3,xname);
            else
               numbreaks = 0;
            break;
         case 'T':  /* trace for n (in hex) instrs, or default 1
                       NOTE:  Unlike debug, Trace always shows the
                       registers before and the instruction about to
                       be executed. */
            flag = breakonio;  /* save status */
            breakonio = FALSE;
            dotrace = 2;
            longnum = 1L;
            instname = DBYTE;
            xname[0] = '\0';
            num3 = 0;
            check_for_csip();
            if (valid)
               {
               if (parse(globalstr))
                  {
                  if (fromhex(globalstr,&longnum))
                     execute(longnum,dotrace,instname,num3,xname);
                  else
                     oops(err_num);
                  }
               else
                  execute(longnum,dotrace,instname,num3,xname);
               }
            breakonio = flag;  /* restore status */
            break;
         case 'W':  /* Where was I? */
            showwhere();
            break;
         case 'R':  /* R for all and current instr, or Rxx for the xx reg */
            if (!parse(globalstr))
               regdisplay();
            else
               {
               if ((regindex = isregname(globalstr)) > -1)
                  {
                  if (!parse(globalstr))
                     {
                     if (regindex != 13)
                        PUTS(regnames[regindex]);
                     else
                        PUTS("IP");
                     PUTS(space_str);
                     PUTS(hexword(&regs.wordarray[regindex],globalstr));
                     getstr("\n:",globalstr);
                     }
                  if (globalstr[0])
                     if (fromhex(globalstr,&temp))
                        {
                        if (globalstr[0])
                           regs.wordarray[regindex] = temp;
                        if ((regindex == 9) || (regindex == 13))  /* CS or IP */
                           {
                           where_count = 0;
                           errmessage = NOHALT;
                           }
                        }
                     else
                        oops(err_num);
                  }
               else
                  {
                  if ((toupper(globalstr[0]) == 'F') && (globalstr[1] == '\0'))
                     modify_flags();
                  else
                     oops("register?\n");
                  }
               }
            break;
         case 'M':  /* do Math */
            if (getnumber(&longnum))
               {
               PUTS(hexword(&longnum,globalstr));
               PUTS(nl_str);
               }
            else
               oops(err_num);
            break;
         case 'D':  /* dump, format is [seg:]startofs [endofs]
                       use ESC to interrupt output */
            flag = BYTE;  /* default to BYTE dump */
            count = cmdptr;  /* set up for DStack check, if necessary */
            dumplen = dumpofs + 0x7f;
            skip_cmd_spaces();
            num1 = 4;  /* mode for getaddrange */
            if (toupper(cmdline[count]) == 'W')
               {
               flag = WORD;  /* do a WORD dump */
               cmdptr = count + 1;  /* and bypass the W for getaddrange() */
               dumplen = dumpofs + 0x3f;  /* half the number of words */
               num1 = 6;  /* new mode for getaddrange */
               }
            if (getaddrange(&dumpseg,&dumpofs,&dumplen,num1))
               {
               dumplen -= dumpofs;
               dump(dumpseg,dumpofs,dumplen,flag);
               dumpofs += dumplen + 1;
               if (flag == WORD)
                  dumpofs += dumplen + 1;
               }
            else
               {
               cmdptr = count;
               skip_cmd_spaces();
               cmdptr++;
               if (toupper(cmdline[cmdptr-1]) == 'S')
                  {
                  count = 14;
                  if (parse(globalstr))
                     if (valid = fromhex(globalstr,&longnum))
                        if (longnum < 0x8000L)
                           count = longnum + longnum - 2;
                  if (!valid)
                     oops(err_num);
                  else
                     {
                     do
                        {
                        holdofs = regs.r.sp + count;
                        num1 = fetchword(regs.r.ss,holdofs);
                        PUTS(hexword(&regs.r.ss,globalstr));
                        PUTS(":");
                        PUTS(hexword(&holdofs,globalstr));
                        PUTS(space_str);
                        PUTS(hexword(&num1,globalstr));
                        PUTS(nl_str);
                        count -= 2;
                        } while ((count != -2) && (gchar() != ESC_KEY));
                     }
                  }
               else
                  oops(err_rng);
               }
            break;
         case 'U':  /* Unasm */
            holdip = regs.r.ip;
            holdcs = regs.r.cs;
            unasmlen = unasmip + 0x1f;  /* not really a length */
            if (getaddrange(&unasmcs,&unasmip,&unasmlen,3))
               {
               regs.r.cs = unasmcs;
               regs.r.ip = unasmip;
               do
                  {
                  do
                     {
                     prefix = 0;
                     tempip = regs.r.ip;
                     decode(hexstr,globalstr);
                     putsinstr(regs.r.cs,tempip,hexstr,globalstr,NO);
                     } while ( ((keyin = gchar()) != ESC_KEY)
                           && (instruction.flags & PREFIX) );
                  } while ((regs.r.ip <= unasmlen) && (keyin != ESC_KEY));
               }
            unasmip = regs.r.ip;
            unasmcs = regs.r.cs;
            regs.r.ip = holdip;
            regs.r.cs = holdcs;
            break;
         case 'A':  /* Asm */
            if (getaddrange(&asmcs,&asmip,&unasmlen,3))
               {
               asmspacing = 9;
               do
                  {
                  PUTS(hexword(&asmcs,globalstr));
                  PUTS(":");
                  PUTS(hexword(&asmip,globalstr));
                  getstr(space_str,cmdline);
                  cmdptr = 0;
                  skip_cmd_spaces();
                  tempip = cmdptr;
                  if (cmdline[cmdptr])
                     if (vimasm())
                        {
                        for (count = 0; count <= bytectr; count++)
                           putbyte(asmcs,asmip+count,xname[count]);
                        asmip += bytectr+1;
                        }
                  } while (cmdline[tempip]);
               }
            asmspacing = 0;
            break;
         case 'F':  /* Fill */
            fillenter(2,1);  /* 2 = addr format, 1 = fill or enter */
            break;
         case 'E':  /* Enter */
            fillenter(1,1);  /* 1 = addr format, 1 = fill or enter */
            break;
         case 'S':  /* Search */
            fillenter(2,0);  /* 2 = addr format, 0 = search */
            break;
         case 'Q':  /* quit program */
         case ';':  /* comment */
            break;
         default:
            if (byte != '\0')  /* if not blank line */
               oops(err_cmd);
            break;
         }
      } while (byte != 'Q');
   }

check_for_csip()
   {
   skip_cmd_spaces();
   if (cmdline[cmdptr] == '=')
      {
      cmdptr++;  /* bypass the '=' */
      if (valid = getaddrange(&regs.r.cs,&regs.r.ip,&temp,1))
         {
         where_count = 0;
         errmessage = NOHALT;
         skip_cmd_spaces();
         }
      }
   }

set_breakpoints(num1,savelist_flag)
   unsigned int num1;  /* number of breakpoints to set */
   int savelist_flag;  /* for B, save list of breakpoints */
   {
   unsigned int holdseg,holdofs,holdofs2,valid2;

   holdseg = regs.r.cs;
   valid2 = TRUE;
   if ( ((valid = getprotrange(&holdseg,&holdofs,&holdofs2)) > 0)
             && (numbreaks < num1))
                        /* set breakpoints, if any */
      do
         {
         numbreaks++;
         templong = holdseg;
         brkpoints[numbreaks].lowaddr = ((templong << 4) + holdofs) & 0xfffffL;
         brkpoints[numbreaks].highaddr = ((templong << 4) + holdofs2) & 0xfffffL;
                                  /* limit break addresses to 20 bits */
         if (cmdline[cmdptr] == '=')  /* if a break count */
            {
            cmdptr++;
            if (valid2 = getnumber(&templong))
               brkcounts[numbreaks] = templong;
            else
               oops(err_rng);
            }
         else
            brkcounts[numbreaks] = 1;
         if (savelist_flag)
            {
            strcat(xname,hexword(&holdseg,globalstr));
            strcat(xname,":");
            strcat(xname,hexword(&holdofs,globalstr));
            if (holdofs2 != holdofs)
               {
               strcat(xname,",");
               strcat(xname,hexword(&holdofs2,globalstr));
               }
            if (brkcounts[numbreaks] != 1)
               {
               strcat(xname,"=");
               strcat(xname,hexword(&brkcounts[numbreaks],globalstr));
               }
            strcat(xname,space_str);
            }
         holdseg = regs.r.cs;
         } while ( (valid2)
             && ((valid = getprotrange(&holdseg,&holdofs,&holdofs2)) > 0)
             && (numbreaks < num1));
   if (valid2 == FALSE)
      valid = FALSE;
   }

parse_z_cmds()
   {
   int ch,numin;
   UNSIGNED_LONG oldpos,t_seq,t_crt,t_attrib,t_si;

   skip_cmd_spaces();
   switch(toupper(cmdline[cmdptr++]))
      {
      case '7':  /* toggle 8087 WAIT instrction status */
         PUTS("WAIT instruction breakpoint ");
         onoroff(wait_status);
         wait_status = !wait_status;
         break;
      case 'C':  /* Toggle PC-XT clone status */
         pclone = !pclone;
         PUTS("PC-clone emulation ");
         onoroff(pclone);
         break;
      case 'S':  /* Toggle screen swapping */
         pcscreen = !pcscreen;
         PUTS("Screen swap ");
         onoroff(pcscreen);
         if (pcscreen)
            {
            checkaddress = checkaddress2;  /* use saved video memory */
            checkwaddress = checkwaddress2;
            }
         else
            {
            checkaddress = checkaddress1;  /* use normal video memory */
            checkwaddress = checkwaddress1;
            }
         break;
      case 'V':  /* fake retrace */
         fake_retrace = !fake_retrace;
         if (parse(globalstr))
            if (fake_retrace = fromhex(globalstr,&longnum))
               {
               retr_count = longnum;
               retr_num = retr_count;
               }
            else
               oops(err_num);
         else
            {
            retr_count = 18;  /* default to 1 second */
            retr_num = 18;
            }
         PUTS("Retrace mode ");
         onoroff(fake_retrace);
         break;
      case 'E':
         skip_cmd_spaces();
         if (toupper(cmdline[cmdptr++]) == 'E')  /* extended mode swapping */
            {
            if (parse(globalstr))  /* if any arguments */
               {
               if ((fromhex(globalstr,&t_seq))
                            && (t_seq >= NORM_SEQ_REG)
                            && (t_seq < NUM_SEQ_REG)
                     && (parse(globalstr)) && (fromhex(globalstr,&t_crt))
                            && (t_crt >= NORM_CRT_REG)
                            && (t_crt < NUM_CRT_REG)
                     && (parse(globalstr)) && (fromhex(globalstr,&t_attrib))
                            && (t_attrib >= NORM_ATTRIB_REG)
                            && (t_attrib < NUM_ATTRIB_REG)
                     && (parse(globalstr)) && (fromhex(globalstr,&t_si))
                            && (t_si >= NORM_SI_REG)
                            && (t_si < NUM_SI_REG))
                  {
                  seq_reg_count = t_seq;
                  si_reg_count = t_si;
                  attrib_reg_count = t_attrib;
                  crt_reg_count = t_crt;
                  extended_swap = TRUE;
                  }
               else
                  {
                  oops(err_num);
                  break;
                  }
               }
            else
               extended_swap = !extended_swap;
            PUTS("Extended swap ");
            onoroff(extended_swap);
            setvideoregs();
            break;
            }
         if (vgascreen)
            {
            vgascreen = -vgascreen;
            PUTS("VGA mode ");
            onoroff(vgascreen);
            setvideoregs();
            if (vgascreen > 0)
               saveegachrs();
            }
         else
            if (egascreen)
               {
               egascreen = -egascreen;
               PUTS("EGA mode ");
               onoroff(egascreen);
               setvideoregs();
               if (egascreen > 0)
                  saveegachrs();
               }
            else
               PUTS("EGA/VGA not present.\n");
         break;
      case 'B':  /* Toggle break-on-io status */
         portbreaknum = 0;  /* empty list */
         valid = TRUE;
         while ((parse(globalstr)) && (valid) && (portbreaknum < 20))
            {
            breakonio = FALSE;
            if (fromhex(globalstr,&longnum))
               portbreaklist[++portbreaknum] = (unsigned int)longnum;
            else
               {
               oops(err_num);
               breakonio = TRUE;
               valid = FALSE;
               }
            }
         breakonio = !breakonio;
         PUTS("Break on I/O ");
         onoroff(breakonio);
         break;
      case 'U':
         int_passing = !int_passing;
         PUTS("Passing of ignored IRQs to system ");
         onoroff(int_passing);
         break;
      case 'Y':
         rom_dma = !rom_dma;
         PUTS("Re-map of ROM DMA during re-mapping of ROM ");
         onoroff(rom_dma);
         break;
      case 'Z':  /* load a "new" ROM! */
         rom_offset = 0;
         rom_count = 0;
         skip_cmd_spaces();
         if (valid = cmdline[cmdptr])
            do
               {
               for (count = 0; (!isspace(cmdline[cmdptr])) && (cmdline[cmdptr]);)
                  globalstr[count++] = cmdline[cmdptr++];
               globalstr[count] = '\0';
/*
puts("globalstr = ");
puts(globalstr);
puts(nl_str);
*/
               if (new_rom())
                  {
/*
puts("cmdline[cmdptr] = ");
puts(hexbyte(&cmdline[cmdptr],xname));
puts(space_str);
puts(&cmdline[cmdptr]);
puts(nl_str);
puts(nl_str);
*/
                  if (parse(globalstr))
                     if ((fromhex(globalstr,&temp))
                          && (((unsigned int)temp & 0xfff) == 0)
                               && ((unsigned int)temp > 0xb000))
                           rom_maps[rom_count++] = temp;
                     else
                        oops("Invalid ROM segment.\n");
                  else
                     rom_maps[rom_count++] = 0xf000;
                  for (count = 0; (int)count < rom_count-1; count++)
                     if (rom_maps[rom_count-1] == rom_maps[count])
                        oops("Duplicate ROM segment.\n");
                  }
               skip_cmd_spaces();
               } while ((cmdline[cmdptr]) && (valid) && (rom_count < 4));
         if (valid == FALSE)
            {
            rom_offset = 0;
            PUTS("Normal ROM restored.\n");
            }
         setmaps();  /* now, fall into load DOS */
      case 'D':  /* start a "local" copy of DOS */
         startdos();
         port64hold = INBPORT(0x64);
         hw8259on = TRUE;
         port60ptr = 0;
         errmessage = NOHALT;  /* clear HALT info */
         for (count = 0; count < 16; count++)
            hwintstack[count] = FALSE;  /* no leftover interrupts */
         hwintflag = FALSE;
         where_count = 0;  /* no instructions in where queue */
         dumpseg = regs.r.ds;  /* defaults for dump and unasm */
         unasmcs =
         asmcs = regs.r.cs;
         dumpofs =
         unasmip =
         asmip = regs.r.ip;
         port_61_save &= 0xfd;  /* make sure speaker is off */
         break;
      case 'R':  /* show video regs */
         showvideoregs();
         break;
      case 'T':  /* timer interrupt divide */
         if (parse(globalstr))
            {
            if (!fromhex(globalstr,&longnum))
               oops(err_num);
            else
               if (parse(globalstr))
                  if ((fromhex(globalstr,&templong)) && (templong > 1)
                       && (templong < 0x10))
                     irq_divisors[templong] = longnum;
                  else
                     oops(err_num);
               else
                  clockdivide = longnum;
            }
         else
            {
            PUTS("Divisors:\nClock ");
            PUTS(hexword(&clockdivide,globalstr));
            PUTS(nl_str);
            PUTS("IRQ ");
            for (count = 2; count < 16; count++)
               {
               PUTS(hexbyte(&count,globalstr));
               PUTS("   ");
               }
            PUTS("\n   ");
            for (count = 2; count < 16; count++)
               {
               PUTS(hexword(&irq_divisors[count],globalstr));
               PUTS(space_str);
               }
            PUTS(nl_str);
            }
         break;
      case 'H':  /* set hotkey */
         if ((!getnumber(&longnum)) || (longnum > 0x7fL)
                    || (longnum == 0x36L))
                       /* do not allow right-shift + right-shift */
            oops(err_num);
         else
            hkscancode = longnum;
         break;
      case 'M':  /* how much memory available */
         disp_mem_size();
         break;
      case 'P':  /* Pre-DOS 3.0 patch location */
         get_patches();
         break;
      case 'N':
         skip_cmd_spaces();
         if (cmdline[cmdptr])
            strcpy(imagename,&cmdline[cmdptr]);
         else
            oops("Image name?\n");
         break;
      case 'L':
         port60ptr = 0;
         for (count = 0; count < 16; count++)
            hwintstack[count] = FALSE;  /* no leftover interrupts */
         hwintflag = FALSE;
         where_count = 0;  /* Forget the "Where was I" information. */
      case 'W':
         rwimg(toupper(cmdline[cmdptr-1]));
         break;
      case 'O':  /* re-direct output */
         skip_cmd_spaces();
         strcpy(outfilename,&cmdline[cmdptr]);
         redirectoutput(outfilename,YES);
         break;
      case 'I':  /* re-direct input */
         skip_cmd_spaces();
         strcpy(infilename,&cmdline[cmdptr]);
         redirectinput(infilename,YES);
         break;
      case 'K':  /* toggle passing of keyboard interrupt */
         keybdoff = !keybdoff;
         PUTS("Keyboard interrupt ");
         onoroff(!keybdoff);
         break;
      case 'G':  /* get directory  Uses zf_screen as DTA and a temp string */
         numin = 0;  /* use as a counter */
         strcpy(tempstr,"*.*");
         if (parse(globalstr))
            {
            strcpy(tempstr,globalstr);
            ch = tempstr[strlen(tempstr)-1];
            if (changedir(globalstr,xname) == 0)   /* to, save */
                                     /* if valid path */
               {
               if (ch != '\\')
                  strcat(tempstr,"\\");
               strcat(tempstr,"*.*");
               }
            else
               if (ch == '\\')
                  strcat(tempstr,"*.*");
            changedir(xname,zf_screen);  /* throw away save path */
            }
         if (find_first() == 0)
            {
            do
               {
               PUTS(&zf_screen[30]);  /* show name */
               if (++numin % 4 == 0)
                  PUTS(nl_str);
               else
                  PUTS(strlen(&zf_screen[30]) > 7 ? tab : "\t\t");
               } while ((find_next() == 0) && (gchar() != ESC_KEY));
            }
         else
            PUTS("No files\n");
         if (numin % 4)
            PUTS(nl_str);
         break;
      case 'F':  /* show file */
         if (parse(globalstr))
            {
            strcpy(zf_name,globalstr);
            zf_pos = 0;
            }
         if ((zf_name[0] == '\0') || ((zf_file = open(zf_name,MODE_READ)) == ERR))
            PUTS("No file\n");
         else
            {
            globalstr[0] = '\0';  /* so Again won't go crazy */
            showpage();
            do
               {
               while ((ch = gchar()) > 255)
                  ;
               ch = toupper(ch);
               switch (ch)
                  {
                  case 'G':   /* Home */
                     zf_pos = 0;
                     showpage();
                     break;
                  case 'O':   /* End */
                     zf_pos = lseek(zf_file,0L,2);  /* seek to EOF */
                     pageup(FALSE);
                     showpage();
                     break;
                  case 'H':   /* PgUp and Up */
                  case 'I':
                     pageup(TRUE);
                     showpage();
                     break;
                  case 'P':   /* PgDn and Down */
                  case 'Q':
                     showpage();
                     break;
                  case 'F':   /* Find text */
                  case 'A':
                     pageup(FALSE);  /* get to top of our page */
                     if (ch == 'F')
                        {
                        clrscr(1);
                        curpos(0,0);
                        getstr("Text? ",globalstr);
                        }
                     if (globalstr[0])
                        {
                        if (ch == 'A')
                           getpage();  /* again goes to next page */
                        strupr(globalstr);
                        oldpos = zf_pos;  /* remeber where we are */
                        while (numin = getpage())
                           {
                           strupr(zf_screen);
                           if (findstring(globalstr,zf_screen) != -1)
                              break;
                           }
                        if (numin == 0)
                           zf_pos = oldpos;
                        else
                           pageup(FALSE);  /* get back to start of found page */
                        }
                     showpage();
                     if (numin == 0)
                        {
                        curpos(1,6);
                        PUTS("Text not found");
                        }
                     break;
                  }
               } while (ch != ESC_KEY);
            pageup(FALSE);  /* return to start of this page */
            close(zf_file);  /* always close upon leaving */
            curpos(24,0);
            PUTS(nl_str);
            }
         break;
      default:
         oops(err_cmd);
         break;
      }
   }

void get_patches()
   {
   unsigned int num1,num2,num3,valid;

   valid = TRUE;
   patched = 0;
   num3 = cmdptr;  /* set up to restore the command line from parsing */
   while ((parse(globalstr)) && (valid) && (patched < 20))
      {
      cmdptr = num3;  /* "unparse" the command line */
      num1 = patchseg[patched];  /* use temporary storage for now */
      num2 = patchofs[patched];
      count = patchbyte[patched];
      if (valid = getaddrange(&num1,&num2,&count,5)) /* get seg:ofs */
         {
         count = PEEK(num1,num2);
         patched++;
         patchseg[patched] = num1;  /* save the patch info */
         patchofs[patched] = num2;
         patchbyte[patched] = (count & 0x0f) + ((maxpara >> 8) - 0x10);
         }
      else
         patched = 0;
      num3 = cmdptr;  /* set up to "unparse" again */
      }
   if (patched)
      {
      for (num1 = 1; num1 <= patched; num1++)
         putbyte(patchseg[num1],patchofs[num1],patchbyte[num1]);
      PUTS("DOS patched.\n");
      }
   else
      oops(err_num);
   }

void skip_cmd_spaces()
   {
   while ((cmdline[cmdptr]) && (isspace(cmdline[cmdptr])))
      cmdptr++;
   }

disp_mem_size()
   {
   unsigned int valid,mem_para;
   UNSIGNED_LONG longnum;

   valid = TRUE;
   if (parse(globalstr))
      {
      valid = fromhex(globalstr,&longnum);
      mem_para = (unsigned int)longnum;
      if ((!valid) || (mem_para & 0xfff) || (mem_para < 0xc000)
                   || (mem_para > 0xf000))
         {
         oops(err_num);
         valid = FALSE;
         }
      else
         {
         deadmem = mem_para;
         setmaps();
         }
      }
   if (valid)
      {
      PUTS("Available paragraphs: ");
      PUTS(hexword(&maxpara,globalstr));
      PUTS("   \"Non-existant\" RAM maps to: ");
      PUTS(hexword(&deadmem,globalstr));
      PUTS(nl_str);
      }
   }

void onoroff(flag)
   int flag;
   {
   if (flag > 0)
      PUTS("enabled.\n");
   else
      PUTS("disabled.\n");
   }

int make_inst_string(fromstr,tostr)
   char *fromstr, *tostr;
   {
   int i,j,flag;

   i = 1;
   j = 1;
   tostr[0] = '\t';
   flag = TRUE;
   while (fromstr[i] != fromstr[0])
      {
      if ((isspace(fromstr[i])) && (flag))
         {
         tostr[j++] = '\t';
         i++;  /* Aztec requirement! */
         while (isspace(fromstr[i]))
            i++;  /* bypass multiple spaces/tabs */
         flag = FALSE;
         }
      else
         tostr[j++] = fromstr[i++];
      }
   tostr[j] = '\0';
   return(j);
   }

void showwhere()
   {
   unsigned int num1,num2,holdcs,holdip,tempip,count,valid,keyin;
   unsigned char hexstr[20];
   UNSIGNED_LONG temp;

   if (where_count)
      {
      temp = 16L;  /* default to 16 instructions */
      valid = TRUE;
      if (parse(globalstr))
         valid = fromhex(globalstr,&temp);
      if (valid)
         {
         if (temp > where_count)  /* if not that many were executed */
            temp = where_count;
         num1 = (unsigned int)temp;
         if (num1 > WHERE_SIZE)
            num1 = WHERE_SIZE;
         num2 = (whereptr + (WHERE_SIZE+1 - num1)) & WHERE_SIZE;  /* get start */
         holdcs = regs.r.cs;  /* save them for a while */
         holdip = regs.r.ip;
         for (count = 0; count < num1; count++)
            {
            regs.r.ip = wherequeueofs[num2];
            regs.r.cs = wherequeueseg[num2];
            prefix = 0;
            do
               {
               tempip = regs.r.ip;
               decode(hexstr,globalstr);
               putsinstr(regs.r.cs,tempip,hexstr,globalstr,NO);
               } while ( ((keyin = gchar()) != ESC_KEY)
                     && (instruction.flags & PREFIX) );
            num2 = (num2 + 1) & WHERE_SIZE;  /* update and wrap */
            if (keyin == ESC_KEY)
               count = num1;
            }
         regs.r.ip = holdip;
         regs.r.cs = holdcs;
         }
      else
         oops(err_num);
      }
   else
      PUTS("No instructions have been executed.\n");
   }


void regdisplay()  /* display registers and current instruction */
   {
   unsigned int temp,temp2,tempip,tip;
   unsigned char name[20],hexstr[20];
   char *hexword();

   dispregs();
   count = 0;
   temp =
   tempip = regs.r.ip;
   temp2 = regs.r.cs;
   prefix = 0;
   do
      {
      tip = regs.r.ip;
      decode(hexstr,globalstr);
      count++;
      putsinstr(regs.r.cs,tip,hexstr,globalstr,YES);
      } while ((instruction.flags & PREFIX) && (gchar() != ESC_KEY)
                        && (count <4));
   PUTS(nl_str);
   regs.r.ip = temp;
   regs.r.cs = temp2;
   }

void putsinstr(seg,ofs,hex,instr,dispmem)
            /* display seg:ofs hex instr [memref] */
   unsigned int seg,ofs,dispmem;
   unsigned char *hex, *instr;
   {
   unsigned char temp[20];

   PUTS(hexword(&seg,temp));
   PUTS(":");
   PUTS(hexword(&ofs,temp));
   PUTS("  ");
   PUTS(hex);
   if (strlen(hex) < 10)
      PUTS("   ");
   PUTS(instr);
   if (dispmem)
      {
      if ( ((instruction.dest == DM) && (mod < 3))
                                || (instruction.dest == DADDR16))
         {
         PUTS(&tabstr[strlen(instr)]);
         PUTS(hexword(&dest.seg,temp));
         PUTS(":");
         PUTS(hexword(&dest.ofs,temp));
         PUTS("=");
         if (instruction.flags & BYTE)
            PUTS(hexbyte(&dest.value,temp));
         else
            PUTS(hexword(&dest.value,temp));
         }
      if ( ((instruction.source == SM) && (mod < 3))
                                  || (instruction.source == SADDR16))
         {
         PUTS(&tabstr[strlen(instr)]);
         PUTS(hexword(&source.seg,temp));
         PUTS(":");
         PUTS(hexword(&source.ofs,temp));
         PUTS("=");
         if (instruction.flags & BYTE)
            PUTS(hexbyte(&source.value,temp));
         else
            PUTS(hexword(&source.value,temp));
         }
      }
   PUTS(nl_str);
   }

void showvideoregs()
   {
   int i;

   if ((egascreen > 0) || (vgascreen > 0))
      {
      PUTS("Misc (3C2): ");
      PUTS(hexbyte(&egaports[P3C2],globalstr));
      if (videobase < 0x3d0)
         PUTS("   Feature (3BA): ");
      else
         PUTS("   Feature (3DA): ");
      PUTS(hexbyte(&egaports[P3xA],globalstr));
      if (vgascreen > 0)
         {
         PUTS("\nVideo subsystem enable (3C3): ");
         PUTS(hexbyte(&egaports[P3C3],globalstr));
         }
      else
         {
         PUTS("\nGraphics I (3CC): ");
         PUTS(hexbyte(&egaports[P3CC],globalstr));
         PUTS("   Graphics II (3CA): ");
         PUTS(hexbyte(&egaports[P3CA],globalstr));
         }
      if (extended_swap)
         {
         PUTS("\n3BF (Hercules port): ");
         PUTS(hexbyte(&mdaports[P3BF],globalstr));
         PUTS("  3CB: ");
         PUTS(hexbyte(&egaports[P3CB],globalstr));
         PUTS("  3CD: ");
         PUTS(hexbyte(&egaports[P3CD],globalstr));
         }
      PUTS("\nSequencer (3C4): ");
      PUTS(hexbyte(&egaports[P3C4],globalstr));
      PUTS(nl_str);
      for (i = 0; i <= seq_reg_count; i++)
         {
         PUTS(hexbyte(&seqregs[i],globalstr));
         PUTS(space_str);
         }
      PUTS("\nCRTC (3");
      if (videobase < 0x3d0)
         {
         PUTS("B4): ");
         PUTS(hexbyte(&mdaports[P3B4],globalstr));
         }
      else
         {
         PUTS("D4): ");
         PUTS(hexbyte(&cgaports[P3D4],globalstr));
         }
      PUTS(nl_str);
      for (i = 0; i <= crt_reg_count; i++)
         {
         if (videobase < 0x3d0)
            PUTS(hexbyte(&mdaregs[i],globalstr));
         else
            PUTS(hexbyte(&cgaregs[i],globalstr));
         PUTS(space_str);
         }
      PUTS("\nAttribute (3C0): ");
      PUTS(hexbyte(&egaports[P3C0],globalstr));
      PUTS("   flip-flop is: ");
      if (attribflipflop)
         PUTS("Data");
      else
         PUTS("Address");
      PUTS(nl_str);
      for (i = 0; i <= attrib_reg_count; i++)
         {
         PUTS(hexbyte(&attribregs[i],globalstr));
         PUTS(space_str);
         }
      PUTS("\nGraphics (3CE): ");
      PUTS(hexbyte(&egaports[P3CE],globalstr));
      PUTS(nl_str);
      for (i = 0; i <= si_reg_count; i++)
         {
         PUTS(hexbyte(&siregs[i],globalstr));
         PUTS(space_str);
         }
      }
   else
      {
      PUTS("CRT: ");
      for (i = 0; i < 18; i++)
         {
         if (videobase < 0x3d0)
            PUTS(hexbyte(&mdaregs[i],globalstr));
         else
            PUTS(hexbyte(&cgaregs[i],globalstr));
         PUTS(space_str);
         }
      if (videobase < 0x3d0)
         {
         PUTS("\nMode (3B8): ");
         PUTS(hexbyte(&mdaports[P3B8],globalstr));
         if (hercules)
            {
            PUTS("    Configuration (3BF): ");
            PUTS(hexbyte(&mdaports[P3BF],globalstr));
            }
         }
      else
         {
         PUTS("\nMode (3D8): ");
         PUTS(hexbyte(&cgaports[P3D8],globalstr));
         PUTS("    Color select (3D9): ");
         PUTS(hexbyte(&cgaports[P3D9],globalstr));
         }
      }
      PUTS(nl_str);
   }

void holdon(hk1,tempip)
   unsigned int hk1,tempip;
   {
   int keyin;

   resetvectors();
   while (gchar() < 256)  /* empty queue before awaiting keypress */
      ;
   while ((keyin = gchar()) > 255)
      ;   /* waste time until a key is pressed */
   if (keyin != ESC_KEY)  /* if not an escape */
      {
      rshiftstatus =
      rshiftused =
      lshiftstatus =
      lshiftused =
      hotkeypressed = FALSE;  /* reset hotkey info */
      if (hk1)  /* if instruction was not executed */
         regs.r.ip = tempip;
      setvectors();
      }
   }

void godoit(prefetch)    /* the Go instruction -- hopefully breakpoints have
                            been set! */
   int prefetch;
   {
   unsigned int tempip,hk1,holdofs;
   unsigned int count;
   int decodeok;
   int basebreaks;
   long bigcount;
   char hexstr[10],*hexlong();

   basebreaks = numbreaks;
   bigcount = 0;
   rshiftstatus =  /* right shift not pressed */
   rshiftused =  /* this right shift keypress has not been used */
   lshiftstatus =  /* left shift not pressed */
   lshiftused =  /* this left shift keypress has not been used */
   hotkeypressed =  /* hot key not activated */
   hk1 = FALSE;  /* hk1 used to say if hotkey pressed during decode
                    or execute */
   OUTBPORT(port_61_save,0x61);  /* restore speaker info */
   if (pcscreen)
      {
      setvimscreen();  /* set up virtual screen */
      checkaddress = checkaddress1;  /* use standard virtual memory */
      checkwaddress = checkwaddress1;
      }
   setvectors();  /* look at interrupts */
   do
      {
      do
         {
         if ((prefetch) && (regs.r.cs != 0xf000))
            {
            keyin = 0;  /* just something besides an ESC */
            numbreaks = 0;  /* do not count breaks while finding
                               the next instruction */
            holdofs = regs.r.ip;
            do
               {
               decodens();  /* update IP.  Strings are not needed */
               } while ((instruction.flags & PREFIX)
                           && ((keyin = gchar()) != ESC_KEY));
            numbreaks = basebreaks + 1;
            templong = regs.r.cs;
            brkpoints[numbreaks].lowaddr =
                           ((templong << 4) + regs.r.ip) & 0xfffffL;
            brkpoints[numbreaks].highaddr = brkpoints[numbreaks].lowaddr+4;
                                           /* +4 for prefetch size! */
            brkcounts[numbreaks] = 1;  /* only one count for a J */
            regs.r.ip = holdofs;
            if (keyin == ESC_KEY)
               memok = FALSE;
            }
         prefixip =   /* global for location of last prefix on int */
         start_ip =   /* global for location of all prefixes */
         tempip = regs.r.ip;  /* local for location of all prefixes, no int */
         prefix = 0;  /* clear out prefixes */
         dest.where =
         source.where = MEMORY;  /* default to MEMORY for source/destination */
         do
            {
            decodeok = decodens();
            } while ((instruction.flags & PREFIX) && (decodeok) && (memok)
                 && (!(hk1 = hotkeypressed)));
/* hk1 is used to show hotkey interrupt during instruction fetch and
   decode (i.e. BEFORE the instruction was executed) */
         if ((decodeok) && (memok) && (!hk1))
            {
            wherequeueofs[whereptr] = tempip;  /* ip for this instruction */
            wherequeueseg[whereptr] = regs.r.cs;  /* cs for this instr */
            (*opfns[instruction.name])();
            numbreaks = basebreaks;  /* in case of prefetch */
            if (memok)
               {
               bigcount++;
               whereptr = (whereptr + 1) & WHERE_SIZE;  /* update and wrap */
               if (prefix & (FORGETZF + ISZFON + ISZFOFF))
                     /* Since they can only be set by a String primitive, they
                        show a REP prefix all by themselves */
                  {
                  regs.r.cx.word--;
                  if (prefix & FORGETZF)
                     regs.r.ip = tempip;
                  else
                     if (prefix & ISZFON)
                        {
                        if (regs.r.flag & ZERO)
                           regs.r.ip = tempip;
                        }
                     else
                        if (!(regs.r.flag & ZERO))
                           regs.r.ip = tempip;
                  }
               if (interruptcount)
                         /* interrupts not allowed due to segreg reference */
                  interruptcount--;
               else
                  tempip = gocheckints();
                     /* Trap interrupt, for example.  The interrupt number is 
                        not displayed by this version of the routine. */
               }
            }
         } while ((decodeok) && (memok) && (!hotkeypressed));
      if (hotkeypressed == 2)
         holdon(hk1,tempip);
      } while ((decodeok) && (memok) && (!hotkeypressed));
   resetvectors();  /* restore interrupts to system when we're done */
   if (pcscreen)
      {
      setrealscreen();  /* and back to the "real" screen */
      checkaddress = checkaddress2;  /* use saved video memory */
      checkwaddress = checkwaddress2;
      }
   while (gchar() < 256)
      ;   /* empty the keyboard buffer */
   port_61_save = INBPORT(0x61);  /* get speaker info */
   OUTBPORT(port_61_save&0xfd,0x61);  /* make sure speaker is off */
   PUTS(nl_str);
   PUTS("Number of instructions executed: ");
   PUTS(hexlong(&bigcount,hexstr));  /* debugging statements */
   PUTS(nl_str);
   where_count += bigcount;
   numbreaks = 0;  /* remove breakpoints so we don't break on them during
                      Unasm, Reg, etc. */
   if ((!decodeok) || (!memok) || (hk1))  /* if instruction not executed */
      regs.r.ip = tempip;  /* go back to start of the instruction */
   memok = TRUE;  /* let the simulator at the byte again */
   regdisplay();  /* show where we quit */
   }

void execute(longnum,dotrace,instname,check_name,inst_name_str)
     /* the eXecute instruction -- execute a given number of instructions. */
   UNSIGNED_LONG longnum;
   unsigned int dotrace,instname,check_name; /* TRUE or FALSE on trace mode,
                                     name of instruction to break on */
   char *inst_name_str;
   {
   unsigned int tip,tempip,hk1;
   unsigned int count;
   int decodeok;
   long bigcount;
   char hexstr[20],*hexlong();

   bigcount = 0;
   rshiftstatus =
   rshiftused =
   lshiftstatus =
   lshiftused =
   hotkeypressed =
   hk1 = FALSE;  /* reset hotkey stuff */
   OUTBPORT(port_61_save,0x61);  /* restore speaker mode */
   if (!dotrace)
      {
      if (pcscreen)  /* if swapping and no trace */
         {
         setvimscreen();  /* set up virtual screen */
         checkaddress = checkaddress1;  /* use normal video memory */
         checkwaddress = checkwaddress1;
         }
      setvectors();  /* and look at interrupts */
      }
   do
      {
      do
         {
         if (dotrace)
             dispregs();
         prefixip =   /* global for location of last prefix on int */
         start_ip =   /* global for location of all prefixes */
         tempip = regs.r.ip;  /* local for location of all prefixes, no int */
         prefix = 0;  /* clear out prefixes */
         dest.where =
         source.where = MEMORY;  /* default to MEMORY for source/destination */
         do
            {
            if ((dotrace) || (check_name))
               {
               tip = regs.r.ip;
               decodeok = decode(hexstr,globalstr);
               if (dotrace)
                  {
                  putsinstr(regs.r.cs,tip,hexstr,globalstr,YES);
                  hotkeypressed = (gchar() == ESC_KEY);
                  }
               if (check_name)
                  {
                  globalstr[check_name] = '\0';
                  if (strcmp(globalstr,inst_name_str) == 0)
                     memok = FALSE;
                  }
               }
            else
               decodeok = decodens();
            } while ((instruction.flags & PREFIX) && (decodeok) && (memok)
                    && (!(hk1 = hotkeypressed)));
   /* hk1 is used to show hotkey interrupt during instruction fetch and
      decode (i.e. BEFORE the instruction was executed) */
         if (dotrace)
            PUTS(nl_str);
         if ((decodeok) && (memok) && (!hk1) && (instruction.name != instname))
            {
            wherequeueofs[whereptr] = tempip;  /* ip for this instruction */
            wherequeueseg[whereptr] = regs.r.cs;  /* cs for this instr */
            (*opfns[instruction.name])();
            if (memok)
               {
               bigcount++;
               whereptr = (whereptr + 1) & WHERE_SIZE;  /* update and wrap */
               if (prefix & (FORGETZF + ISZFON + ISZFOFF))
                     /* Since they can only be set by a String primitive, they
                        show a REP prefix all by themselves */
                  {
                  regs.r.cx.word--;
                  if (prefix & FORGETZF)
                     regs.r.ip = tempip;
                  else
                     if (prefix & ISZFON)
                        {
                        if (regs.r.flag & ZERO)
                           regs.r.ip = tempip;
                        }
                     else
                        if (!(regs.r.flag & ZERO))
                           regs.r.ip = tempip;
                  }
               if (interruptcount)  /* interrupts not allowed due to segreg reference */
                  interruptcount--;
               else
                  if (dotrace)
                     tempip = checkints();   /* trap interrupt, for example */
                  else
                     tempip = gocheckints();  /* do not tell of interrupts */
               }
            }
         if (dotrace)
            if (gchar() == ESC_KEY)
               hotkeypressed = 1;
         } while ((decodeok) && (memok) && (bigcount < longnum)
               && (instruction.name != instname) && (!hotkeypressed));
      if (hotkeypressed == 2)
         holdon(hk1,tempip);
      } while ((decodeok) && (memok) && (bigcount < longnum)
            && (instruction.name != instname) && (!hotkeypressed));
   if (!dotrace)
      {
      resetvectors();  /* restore interrupts to system when we're done */
      if (pcscreen)
         {
         setrealscreen();  /* and back to the "real" screen */
         checkaddress = checkaddress2;  /* use saved video memory */
         checkwaddress = checkwaddress2;
         }
      }
   while (gchar() < 256)
      ;  /* clear the keyboard queue */
   port_61_save = INBPORT(0x61);  /* get speaker info */
   OUTBPORT(port_61_save&0xfd,0x61);  /* make sure speaker is off */
   where_count += bigcount;
   numbreaks = 0;  /* remove breakpoints so we don't break on them during
                      Unasm, Reg, etc. */
   if ((!decodeok) || (!memok) || (hk1) || (instruction.name == instname))
                      /* if instruction not executed */
      regs.r.ip = tempip;  /* go back to start of the instruction */
   memok = TRUE;  /* let the simulator at the byte again */
   if ((dotrace != 2) || ((bigcount != longnum) && (longnum != 1L)))
      {
      PUTS(nl_str);
      PUTS("Number of instructions executed: ");
      PUTS(hexlong(&bigcount,hexstr));  /* debugging statements */
      PUTS(nl_str);
      regdisplay();  /* show where we quit */
      }
   }

int checkints()  /* Note:  NMI, single step (trap), and software interrupts
         all have priority over hardware interrupts (INTR pin).  Of these
         3 non-maskable interrupts, when 2 occur simultaneously, the
         priority order is (highest to lowest): single step (trap), NMI, and
         software interrupts.  However, when 3 of them occur simultaneously,
         the order is NMI, software interrupts, and then single step (trap).

         Sigh... If you want to get the order right, have fun.  I'm going to
         leave them in this order.

         Note the absence of NMI handling.  See the comments at the top of
         the program. */
   {
   int i,j,k,l;  /* used to check for next interrupt */
   unsigned int saveip,inst_name;

   saveip = regs.r.ip;  /* save in case INT fails */
   if (memok)
      {
      if (regs.r.flag & TRAP)
         {
         inst_name = instruction.name;  /* save for break on name INT */
         instruction.name = INT;  /* do a TRAP interrupt */
         instruction.flags = 0;
         source.value = 1L;  /* INT 1 is trap */
         if (prefix & (FORGETZF + ISZFON + ISZFOFF))  /* if in a string
                                                         primitive */
            regs.r.ip = prefixip;
         prefix = 0;
         PUTS("Doing a TRAP interrupt\n\n");
         (*opfns[instruction.name])();  /* do the INT instruction */
         instruction.name = inst_name;  /* restore for break on name INT */
         if (memok == FALSE)
            {
            regs.r.ip = saveip;
            return(regs.r.ip);
            }
         }
      else
                   /* hardware interrupts are less common, check them first */
         if ((hwintflag) && (hw8259on) && (regs.r.flag & INTERRUPT) &&
               ((hwintstack[0]) || (!hwintstack[1]) || (!eoiwait)))
            {
            hwintnumber = -1;  /* set to default not found */
            k = -1;  /* set k to default no pending interrupt */
            for (i = 15; i >= 0; i--)
                   /* Check in reverse order, so low numbers have
                      priority */
               if (hwintstack[i])  /* if pending interrupt, note it */
                  {
                  k = hwintnumber;
                            /* hold record of more than 1 pending Int */
                  hwintnumber = hwintstack[i];
                  l = i;
                  }
            if (errmessage == HALTERR)
               {
               regs.r.ip++;  /* point to instruction after the HLT */
               errmessage = NOHALT;  /* reset flag */
               }
            inst_name = instruction.name;  /* save for break on name INT */
            instruction.name = INT;  /* do a general interrupt */
            instruction.flags = 0;
            source.value = (UNSIGNED_LONG)hwintnumber;
                           /* the number is... */
            if (hwintnumber == 9)  /* if a keyboard interrupt */
               gchar();  /* remove the key from the regular queue */

            PUTS("\nDoing hardware interrupt # ");
            PUTS(hexbyte(&hwintnumber,globalstr));
            PUTS(nl_str);  /* say who is doing it */

            if (prefix & (FORGETZF + ISZFON + ISZFOFF))
                                       /* if in a string primitive */
               regs.r.ip = prefixip;  /* forget all but the last prefix
                                       (8088 quirk) */
            prefix = 0;
            (*opfns[instruction.name])();  /* go do the INT */
            instruction.name = inst_name;  /* restore for break on name INT */
            if (memok)
               {
               hwintstack[l] = 0;  /* turn off pending flag */
               if (k > -1)  /* if another is waiting */
                  hwintflag = TRUE;  /* alert to pending interrupt */
               else
                  hwintflag = FALSE;  /* clear for next interrupt */
               hw8259on = FALSE;  /* no hardware Ints until turned on */
               }
            else
               {
               regs.r.ip = saveip;
               return(regs.r.ip);
               }
            }
      }
   }


int gocheckints()  /* A version of checkints for the G command.  The
                       comments from checkints() above apply. */
   {
   int i,j,k,l;  /* used to check for next interrupt */
   unsigned int saveip,inst_name;

   saveip = regs.r.ip;  /* save in case INT fails */
   if (memok)
      {
      if (regs.r.flag & TRAP)
         {
         inst_name = instruction.name;  /* save for break on name INT */
         instruction.name = INT;  /* do a TRAP interrupt */
         instruction.flags = 0;
         source.value = 1L;  /* INT 1 is trap */
         if (prefix & (FORGETZF + ISZFON + ISZFOFF))  /* if in a string
                                                         primitive */
            regs.r.ip = prefixip;
         prefix = 0;
         (*opfns[instruction.name])();  /* do the INT instruction */
         instruction.name = inst_name;  /* restore for break on name INT */
         if (memok == FALSE)
            {
            regs.r.ip = saveip;
            return(regs.r.ip);
            }
         }
      else
                   /* hardware interrupts are less common, check them first */
/*         if ((hwintflag) && (hw8259on) && (regs.r.flag & INTERRUPT) &&
               ((hwintstack[0]) || (!hwintstack[1]) || (!eoiwait))) */
         if ((hwintflag) && (hw8259on) && (regs.r.flag & INTERRUPT))
            {
            hwintnumber = -1;  /* set j to default not found */
            k = -1;  /* set k to default no pending interrupt */
            for (i = 15; i >= 0; i--)
                   /* Check in reverse order, so low numbers have
                      priority */
/*               if ((hwintstack[i]) && ((i != 1) || (!eoiwait))) */
               if (hwintstack[i])
                                  /* if pending interrupt, note it */
                  {
                  k = hwintnumber;
                            /* hold record of more than 1 pending Int */
                  hwintnumber = hwintstack[i];
                  l = i;
                  }
            if (errmessage == HALTERR)
               {
               regs.r.ip++;  /* point to instruction after the HLT */
               errmessage = NOHALT;  /* reset flag */
               }
            inst_name = instruction.name;  /* save for break on name INT */
            instruction.name = INT;  /* do a general interrupt */
            instruction.flags = 0;
            source.value = (UNSIGNED_LONG)hwintnumber;
                           /* the number is... */
            if (hwintnumber == 9)  /* if a keyboard interrupt */
               gchar();  /* remove the key from the regular queue */

/*          PUTS("\nDoing hardware interrupt # ");
            PUTS(hexbyte(&hwintnumber,globalstr));
            PUTS(nl_str);  debugging code */

           if (prefix & (FORGETZF + ISZFON + ISZFOFF))
                                       /* if in a string primitive */
               regs.r.ip = prefixip;  /* forget all but the last prefix
                                       (8088 quirk) */
            prefix = 0;
            (*opfns[instruction.name])();  /* go do the INT */
            instruction.name = inst_name;  /* restore for break on name INT */
            if (memok)
               {
               hwintstack[l] = 0;  /* turn off pending flag */
               if (k > -1)  /* if another is waiting */
                  hwintflag = TRUE;  /* alert to pending interrupt */
               else
                  hwintflag = FALSE;  /* clear for next interrupt */
               hw8259on = FALSE;  /* no hardware Ints until turned on */
               }
            else
               {
               regs.r.ip = saveip;
               return(regs.r.ip);
               }
            }
      }
   }

int ispcorat()
   {
   int returnvalue;
/*
   Machine and compiler specific routine to determine if this program
   is running on an 8088/8086 or 80286.  If an 80286 is found, the
   clock interrupt simulation count will be adjusted for the faster
   interpreter.
*/
#asm
   push  sp     ;That's right, push SP!
   pop   bx     ;bx = sp from push.  An 8086/8 will decrement SP, then
                ;perform the push, while the 80286 will push SP and then
                ;decrement.  This is a standard method of determining
                ;which chip is being used.
   sub   bx,sp  ;get difference
   mov   word ptr [bp-2],bx  ;store result
#endasm
   if (returnvalue)
      return(PC);
   return(AT);
   }

void readint(holder,intnum)  /* read interrupt vector intnum into the structure
                               pointed to by holder */
   struct intrptr *holder;
   unsigned int intnum;
   {
   unsigned int offset;

   offset = intnum << 2;  /* offset = intnum * 4 */
   holder->ptrofs = wpeek(0,offset);
   holder->ptrseg = wpeek(0,offset+2);
   }
